"""
Place to store utility functions.
"""
from typing import Dict, Any, Literal, Optional, Union

import numpy as np
import yaml


def read_yaml_file(file_path: str) -> Dict[str, Any]:
    """
    Reads the yaml file and returns as a dictionary.

    :param file_path: Path to yaml file
    :type file_path: str
    :return: Content in the yaml file
    :rtype: dict
    """
    with open(file_path, 'r') as yaml_file:
        data = yaml.load(yaml_file, Loader=yaml.FullLoader)
    return data


def get_nsamples(header: dict) -> int:
    """
    Takes header dictionary, gets all nsamples.
    Finds the mostly occurring number and returns.
    
    :param header: Header information of all loaded time series.
    :type header: dict
    :return: Most common number in nsamples
    """
    nsamples_list = []

    for key, sub_dict in header.items():
        nsamples = sub_dict.get('nsamples')
        if nsamples is not None:
            nsamples_list.append(nsamples)
        else:
            # If 'nsamples' key is not found in the sub-dictionary, check its values
            for sub_key, value in sub_dict.items():
                if sub_key not in ['Rx', 'Ry'] and isinstance(value, dict):
                    nested_nsamples = value.get('nsamples')
                    if nested_nsamples is not None:
                        nsamples_list.append(nested_nsamples)

    most_common_number = max(set(nsamples_list), key=nsamples_list.count)
    return most_common_number


def get_fftlength(nofsamples: int) -> int:
    """
    Returns a suggested FFT length based on the no. of samples.
    :param nofsamples: No. of samples
    :type nofsamples: int
    :return: Suggested FFT length
    :rtype: int
    """
    # Based on Borah & Patro, 2015
    i = 1
    ffts = [256]
    cfft = 256 * (2 ** i)
    term = nofsamples / (20 * i);
    ffts.append(cfft)
    while cfft <= term:
        i = i + 1
        cfft = 256 * (2 ** i)
        term = nofsamples / (20 * i)
        if cfft <= term:
            ffts.append(cfft)
    window_length = ffts[-1]
    if window_length > 65536:
        window_length = 65536
    return window_length


def get_parzen(fs: float) -> float:
    """
    Returns a suggested parzen window radius based on the sampling frequency.

    :param fs: Sampling frequency
    :type fs: float
    :return: Parzen window radius
    :rtype: float
    """
    if fs >= 512:
        parzen_radius = 0.25
    elif fs > 32:
        parzen_radius = 0.4
    else:  # fs <= 32
        parzen_radius = 0.9
    return parzen_radius


def reshape_array_with_overlap(window_length: int, overlap: int, data: np.ndarray) -> np.ndarray:
    """
    To reshape a numpy array into different windows with 50% overlap.

    :param window_length: Window length that arrays to be divided.
    :type window_length: int
    :param overlap: Overlap required for the reshaping.
    :type overlap: int
    :param data: The data array (1D)
    :type data: np.ndarray

    :return: 2D numpy array, rows= no. of windows, columns = window length
    :rtype: np.ndarray
    """
    # Parameters
    overlap = int(window_length * (overlap / 100))

    # Calculate the number of windows
    num_windows = (len(data) - overlap) // (window_length - overlap)

    # Initialize the 2D array
    result_array = np.empty((num_windows, window_length))

    # Extract windows
    for i in range(num_windows):
        start_idx = i * (window_length - overlap)
        end_idx = start_idx + window_length
        result_array[i] = data[start_idx:end_idx]

    result_array = result_array.T

    return result_array


def get_target_frequency_list(sampling_frequency: float,
                              parzen_window_radius: float,
                              fft_length: int,
                              table_type: Literal['default', 'metronix', 'explicit'] = 'default',
                              frequencies_per_decade: int = 12,
                              lowest_frequency: int = -5,
                              highest_frequency: int = 5,
                              explicit_table: Optional[Union[list, np.ndarray]] = None,
                              ) -> np.ndarray:
    """
    Computes a list of target frequencies based on the given processing parameters.

    :param sampling_frequency: Sampling frequency of the measurement (Hz).
    :type sampling_frequency: float
    :param parzen_window_radius: Radius of the Parzen window.
    :type parzen_window_radius: float
    :param fft_length: Length of the FFT (Fast Fourier Transform).
    :type fft_length: int
    :param table_type: Specifies the type of frequency table to generate.
                       - 'default': Generates the default SigMT frequency table based on `np.logspace`, taking
                                    the frequencies_per_decade, lowest_frequency and highest_frequency parameters.
                       - 'metronix': Generates the Metronix frequency table.
                       - 'explicit': Uses a user-provided explicit list of frequencies.
    :type table_type: Literal['default', 'metronix', 'explicit']
    :param frequencies_per_decade: Number of period/frequency points per decade. This works only for 'default'
                                   table_type.
    :type frequencies_per_decade: int
    :param lowest_frequency: Exponent for the start of the period range (10^start_period).
                         Default is -5, corresponding to 10⁻⁵ Hz. This works only for 'default' table_type.
    :type lowest_frequency: int
    :param highest_frequency: Exponent for the end of the period range (10^stop_period).
                        Default is 5, corresponding to 10⁵ Hz. This works only for 'default' table_type.
    :type highest_frequency: int
    :param explicit_table: A user-provided list or numpy array of target frequencies.
                           This parameter is used only when `table_type` is set to `explicit`.
    :type explicit_table: Optional[Union[list, np.ndarray]]

    :returns: A numpy array (1D) of float (shape: n, ) which is a list of target frequencies.
    :rtype: np.ndarray

    """

    if table_type == 'default':
        frequency_table = np.flip(np.logspace(lowest_frequency, highest_frequency,
                                              int((highest_frequency - lowest_frequency) * frequencies_per_decade + 1)))
    elif table_type == 'explicit':
        if explicit_table is None:
            raise ValueError("explicit_table must be provided when table_type is 'explicit'")
        frequency_table = np.array(explicit_table, dtype=float)
    elif table_type == 'metronix':
        # TODO: Implement this in next version.
        print('Currently Metronix frequency table is not implemented. Switching to default.')
        frequency_table = np.flip(np.logspace(lowest_frequency, highest_frequency,
                                              int((highest_frequency - lowest_frequency) * frequencies_per_decade + 1)))
    else:
        print('table_type provided is not supported. Switching to default.')
        frequency_table = np.flip(np.logspace(lowest_frequency, highest_frequency,
                                              int((highest_frequency - lowest_frequency) * frequencies_per_decade + 1)))

    fr = parzen_window_radius * frequency_table  # bandwidth of parzen window - oneside from ft
    total_bandwidth = fr * 2  # bandwidth of parzen window - two side from ft

    # No. of spectral bins in the parzen window for each ft
    dof = total_bandwidth / (sampling_frequency / fft_length)

    maximum = sampling_frequency / 2

    # Hardcoding maximum frequency to 15000
    if maximum > 15000:
        maximum = 15000
    f_max = max(frequency_table[frequency_table < maximum])  # ft_max, following nyquist
    f_max_index = np.where(frequency_table == f_max)[0][0]  # index in ftable

    # Finding minimum frequency bin used for a parzen window used.
    parzen_lower_extend = frequency_table - (frequency_table * parzen_window_radius)
    # Find the index of the target frequency whose lower Parzen window boundary is
    # closest to the frequency resolution
    f_min_index = np.argmin(abs(parzen_lower_extend - (sampling_frequency / fft_length)))

    # Looks for minimum dof=10
    f_min_index = np.min([f_min_index, np.argmin(abs(dof-10))])

    ft_list = frequency_table[f_max_index:f_min_index]  # Making ft_list

    return ft_list



"""
Functions to plot output curves
"""

import matplotlib
import numpy as np
import xarray as xr
from matplotlib import pyplot as plt

matplotlib.use("Qt5Agg")
plt.rcParams['figure.max_open_warning'] = 50


def plot_mt_app_res(dataset: xr.Dataset, procinfo: dict) -> None:
    """
    Plots apparent resistivity and phase

    :param dataset: Dataset containing estimates
    :type dataset: xr.Dataset
    :param procinfo: Dictionary containing processing information
    :type procinfo: dict
    :return: None
    :rtype: None

    """
    if not dataset:
        raise ValueError('dataset invalid')

    # Apparant resistivities and phase
    ftlist = dataset['frequency'].values
    zxy = dataset['zxy'].values
    zyx = dataset['zyx'].values
    zxy_var = dataset['zxy_var'].values
    zyx_var = dataset['zyx_var'].values

    rho_xy = (0.2 / ftlist) * (abs(zxy) ** 2)
    rho_yx = (0.2 / ftlist) * (abs(zyx) ** 2)
    phase_xy = np.degrees(np.arctan(zxy.imag / zxy.real))
    phase_yx = np.degrees(np.arctan(zyx.imag / zyx.real))
    #
    #
    # Errors for app. resistivity and phase
    err_rxy = (0.4 / ftlist) * abs(zxy) * zxy_var
    err_ryx = (0.4 / ftlist) * abs(zyx) * zyx_var
    err_pxy = np.degrees(zxy_var / abs(zxy))
    err_pyx = np.degrees(zyx_var / abs(zyx))
    #
    # Plotting section
    # Plot App. res & Phase
    plt.figure()
    plt.subplot(211)
    plt.suptitle(f'{procinfo["localsite"]} - {procinfo["fs"]} Hz')
    plt.scatter(ftlist, rho_xy, c='r', s=10, label='XY')
    plt.scatter(ftlist, rho_yx, c='b', s=10, label='YX')
    plt.errorbar(ftlist, rho_xy, yerr=err_rxy, ecolor='r', fmt="none")
    plt.errorbar(ftlist, rho_yx, yerr=err_ryx, ecolor='b', fmt="none")
    plt.xscale('log')
    plt.yscale('log')
    if max(ftlist) < 15000 and min(ftlist) > 0.0001:
        plt.xlim((15000, 0.0001))
    else:
        plt.xlim((max(ftlist) + 10, min(ftlist) - 10))
        if min(ftlist) > 0.001:
            plt.xlim((max(ftlist) + 10, 0.001))
    ymax_magnitude = int(np.ceil(np.log10(max(max(rho_xy), max(rho_yx)))))
    ymin_magnitude = int(np.floor(np.log10(min(min(rho_xy), min(rho_yx)))))
    plt.ylim(10 ** (ymin_magnitude - 2), 10 ** (ymax_magnitude + 2))
    plt.yticks([10 ** i for i in range(ymin_magnitude - 2, ymax_magnitude + 3)])
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('App. Res. (Ohm.m.)')
    plt.legend()
    plt.grid(which='both', linestyle='-.', linewidth=0.4)
    plt.subplot(212)
    plt.scatter(ftlist, phase_xy, c='r', s=10)
    plt.scatter(ftlist, phase_yx, c='b', s=10)
    plt.errorbar(ftlist, phase_xy, yerr=err_pxy, ecolor='r', fmt="none")
    plt.errorbar(ftlist, phase_yx, yerr=err_pyx, ecolor='b', fmt="none")
    plt.xscale('log')
    if max(ftlist) < 15000 and min(ftlist) > 0.0001:
        plt.xlim((15000, 0.0001))
    else:
        plt.xlim((max(ftlist) + 10, min(ftlist) - 10))
        if min(ftlist) > 0.001:
            plt.xlim((max(ftlist) + 10, 0.001))
    if (min(phase_xy) > 0 and min(phase_yx) > 0) and (max(phase_xy) < 90 and max(phase_yx) < 90):
        plt.ylim((0, 90))
        plt.yticks([0, 15, 30, 45, 60, 75, 90])
    else:
        ylim1 = min(min(phase_xy), min(phase_yx))
        ylim2 = max(max(phase_xy), max(phase_yx))
        plt.ylim(ylim1 - 10, ylim2 + 10)
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Phase (Deg.)')
    plt.grid(which='both', linestyle='-.', linewidth=0.4)
    plt.show()


def plot_tipper(dataset: xr.Dataset, procinfo: dict) -> None:
    """
    Plots tipper

    :param dataset: Dataset containing estimates
    :type dataset: xr.Dataset
    :param procinfo: Dictionary containing processing information
    :type procinfo: dict
    :return: None
    :rtype: None

    """
    if not dataset:
        raise ValueError('dataset invalid')

    ftlist = dataset['frequency'].values
    tzx = dataset['tzx'].values
    txy = dataset['tzy'].values
    #
    tx_a = np.sqrt((np.real(tzx) ** 2) + (np.imag(tzx) ** 2))
    ty_a = np.sqrt((np.real(txy) ** 2) + (np.imag(txy) ** 2))
    tx_p = np.degrees(np.arctan2(tzx.imag, tzx.real))
    ty_p = np.degrees(np.arctan2(txy.imag, txy.real))
    # Plotting section
    plt.figure()
    plt.subplot(211)
    plt.suptitle(f'{procinfo["localsite"]} - {procinfo["fs"]} Hz')
    plt.scatter(ftlist, tx_a, c='r', s=10, label='Tx')
    plt.scatter(ftlist, ty_a, c='b', s=10, label='Ty')
    plt.ylim(0, 1)
    plt.xscale('log')
    if max(ftlist) < 15000 and min(ftlist) > 0.0001:
        plt.xlim((15000, 0.0001))
    else:
        plt.xlim((max(ftlist) + 10, min(ftlist) - 10))
        if min(ftlist) > 0.001:
            plt.xlim((max(ftlist) + 10, 0.001))
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Tipper Amplitude')
    plt.grid(which='both', linestyle='-.', linewidth=0.4)
    plt.legend()
    plt.subplot(212)
    plt.scatter(ftlist, tx_p, c='r', s=10)
    plt.scatter(ftlist, ty_p, c='b', s=10)
    plt.xscale('log')
    if max(ftlist) < 15000 and min(ftlist) > 0.0001:
        plt.xlim((15000, 0.0001))
    else:
        plt.xlim((max(ftlist) + 10, min(ftlist) - 10))
        if min(ftlist) > 0.001:
            plt.xlim((max(ftlist) + 10, 0.001))
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Tipper Phase')
    plt.grid(which='both', linestyle='-.', linewidth=0.4)
    plt.show()


def plot_coherency(dataset: xr.Dataset, procinfo: dict) -> None:
    """
    Plots coherencies

    :param dataset: Dataset containing estimates
    :type dataset: xr.Dataset
    :param procinfo: Dictionary containing processing information
    :type procinfo: dict
    :return: None
    :rtype: None

    """
    if not dataset:
        raise ValueError('dataset invalid')

    plt.figure()
    frequency_values = dataset['frequency'].values
    if 'coh_ex' in dataset:
        plt.scatter(frequency_values, dataset['coh_ex'].values, c='r', label='Ex')
    if 'coh_ey' in dataset:
        plt.scatter(frequency_values, dataset['coh_ey'].values, c='b', label='Ey')
    if 'coh_hz' in dataset:
        plt.scatter(frequency_values, dataset['coh_hz'].values, c='g', label='Hz')
    plt.xscale('log')
    plt.xlim((10000, 0.0001))
    plt.ylim(0, 1)
    # generate tick values for y
    y_ticks = np.arange(0.1, 1.1, 0.1)
    plt.yticks(y_ticks)
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Predicted Coherency')
    plt.title(f'{procinfo["localsite"]} - {procinfo["fs"]} Hz')
    plt.grid(which='both', linestyle='-.', linewidth=0.4)
    plt.legend()
    plt.show()

def plot_coherencies_all(dataset: xr.Dataset) -> None:
    """
    Plots coherency values for all frequencies

    :param dataset: Dataset containing band averaged values
    :type dataset: xr.Dataset
    :return: None
    :rtype: None

    """
    if not dataset:
        raise ValueError('dataset invalid')

    custom_colormap = _get_custom_colormap()

    num_coh = 0
    coh_keys = []
    if 'coh_ex' in dataset:
        num_coh = num_coh + 1
        coh_keys.append('coh_ex')
    if 'coh_ey' in dataset:
        num_coh = num_coh + 1
        coh_keys.append('coh_ey')
    if 'coh_hz' in dataset:
        num_coh = num_coh + 1
        coh_keys.append('coh_hz')
    #
    for freq in dataset['frequency']:
        # Select data for the current frequency
        data_for_freq = dataset.sel(frequency=freq)
        if num_coh == 1:
            tf_component = _get_tf_key(coh_keys, 0)
            plt.figure()
            plt.scatter(data_for_freq[tf_component].real, data_for_freq[tf_component].imag, c=data_for_freq[coh_keys[0]], cmap=custom_colormap,
                        vmin=0, vmax=1)
            plt.colorbar(label=coh_keys[0])  # Add color bar for reference
            plt.title(f"Coherency ({coh_keys[0]}) plot for {freq.values} Hz", fontsize=12)
            plt.xlabel(f'Real ({tf_component})')
            plt.ylabel(f'Imag ({tf_component})')
        elif num_coh == 2:
            fig, axes = plt.subplots(1, 2, figsize=(10, 5))
            for i in range(2):
                tf_component = _get_tf_key(coh_keys, i)
                sc = axes[i].scatter(data_for_freq[tf_component].real, data_for_freq[tf_component].imag, c=data_for_freq[coh_keys[i]],
                                     cmap=custom_colormap, vmin=0, vmax=1)
                fig.colorbar(sc, ax=axes[i], label=coh_keys[i])
                axes[i].set_title(coh_keys[i])
                axes[i].set_xlabel(f'Real ({tf_component})')
                axes[i].set_ylabel(f'Imag ({tf_component})')
            fig.suptitle(f"Coherency plots for {freq.values} Hz", fontsize=12)
        elif num_coh == 3:
            fig, axes = plt.subplots(1, 3, figsize=(15, 5))  # 3 horizontal subplots
            for i in range(3):
                tf_component = _get_tf_key(coh_keys, i)
                sc = axes[i].scatter(data_for_freq[tf_component].real, data_for_freq[tf_component].imag, c=data_for_freq[coh_keys[i]],
                                     cmap=custom_colormap, vmin=0, vmax=1)
                fig.colorbar(sc, ax=axes[i], label=coh_keys[i])
                axes[i].set_title(coh_keys[i])
                axes[i].set_xlabel(f'Real ({tf_component})')
                axes[i].set_ylabel(f'Imag ({tf_component})')
            fig.suptitle(f"Coherency plots for {freq.values} Hz", fontsize=12)

        plt.tight_layout()
        plt.show()


def _get_tf_key(coh_keys, i) -> str:
    """
    Get transfer function key.

    """
    key = ""
    if coh_keys[i] == 'coh_ex':
        key = 'zxy'
    if coh_keys[i] == 'coh_ey':
        key = 'zyx'
    if coh_keys[i] == 'coh_hz':
        key = 'tzx'
    return key

def plot_pd_all(dataset: xr.Dataset) -> None:
    """
    Plots polarization directions values for all frequencies

    :param dataset: Dataset containing band averaged values
    :type dataset: xr.Dataset
    :return: None
    :rtype: None

    """
    if not dataset:
        raise ValueError('dataset invalid')

    afont = {'fontname': 'Arial'}
    yticks = np.arange(-90, 100, 20)

    for freq in dataset['frequency']:
        data_for_freq = dataset.sel(frequency=freq)
        time_array = np.arange(data_for_freq.coords['time_window'].size)
        plt.figure()

        plt.subplot(211)
        plt.suptitle(f"Polarization directions for {freq.values} Hz", fontsize=12)
        plt.scatter(time_array, data_for_freq['alpha_h'].values)
        plt.ylim(-90, 90)
        plt.ylabel(r'$\alpha_H$ (deg.)')
        plt.xticks(**afont, fontsize=12)
        plt.yticks(yticks, **afont, fontsize=12)

        plt.subplot(212)
        plt.scatter(time_array, data_for_freq['alpha_e'].values)
        plt.ylim(-90, 90)
        plt.ylabel(r'$\alpha_E$ (deg.)')
        plt.xlabel('Time window')
        plt.xticks(**afont, fontsize=12)
        plt.yticks(yticks, **afont, fontsize=12)

        plt.show(block=False)

def _get_custom_colormap() -> matplotlib.colors.LinearSegmentedColormap:
    """
       Creates and returns a custom colormap for coherency plots

       :return: Custom colormap
    """
    color_dictionary = {'red': [(0.0, 0.0, 0.0),
                     (0.1, 0.5, 0.5),
                     (0.2, 0.0, 0.0),
                     (0.4, 0.2, 0.2),
                     (0.6, 0.0, 0.0),
                     (0.8, 1.0, 1.0),
                     (1.0, 1.0, 1.0)],
             'green': [(0.0, 0.0, 0.0),
                       (0.1, 0.0, 0.0),
                       (0.2, 0.0, 0.0),
                       (0.4, 1.0, 1.0),
                       (0.6, 1.0, 1.0),
                       (0.8, 1.0, 1.0),
                       (1.0, 0.0, 0.0)],
             'blue': [(0.0, 0.0, 0.0),
                      (0.1, 0.5, 0.5),
                      (0.2, 1.0, 1.0),
                      (0.4, 1.0, 1.0),
                      (0.6, 0.0, 0.0),
                      (0.8, 0.0, 0.0),
                      (1.0, 0.0, 0.0)]}
    return matplotlib.colors.LinearSegmentedColormap('my_colormap', color_dictionary, 256)


"""
The design of all dialog boxes appears from the metronix main window
"""
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QListWidget, QVBoxLayout, QDialogButtonBox, QGroupBox, QHBoxLayout, QLabel, \
    QLineEdit, QComboBox, QGridLayout


class SelectionDialog(QDialog):
    """
    There are cases where user needs to manually select the remote measurement. This class
    is the design of user interface to manually select the measurement for remote reference.
    It shows two boxes with all measurements. User can select the measurements and submit.
    """

    def __init__(self, local_meas, remote_meas, parent=None):
        """
        Constructor
        :param local_meas: Measurements in the local site
        :type local_meas: list
        :param remote_meas: Measurements in the remote site
        :type remote_meas: list
        :param parent: Parent
        :type parent: object

        """
        super().__init__(parent)

        self.setWindowTitle("Select corresponding single and remote measurements")

        # Create two list widgets within group boxes
        self.local_meas_groupbox = QGroupBox("Local Measurements")
        self.local_meas_list = QListWidget()
        self.local_meas_groupbox.setLayout(QVBoxLayout())
        self.local_meas_groupbox.layout().addWidget(self.local_meas_list)

        self.remote_meas_groupbox = QGroupBox("Remote Measurements")
        self.remote_meas_list = QListWidget()
        self.remote_meas_groupbox.setLayout(QVBoxLayout())
        self.remote_meas_groupbox.layout().addWidget(self.remote_meas_list)

        for name in local_meas:
            self.local_meas_list.addItem(name)
        for name in remote_meas:
            self.remote_meas_list.addItem(name)

        self.button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        # Set up the layout
        layout = QVBoxLayout()
        hbox = QHBoxLayout()
        hbox.addWidget(self.local_meas_groupbox)
        hbox.addWidget(self.remote_meas_groupbox)
        layout.addLayout(hbox)
        layout.addWidget(self.button_box)

        self.setLayout(layout)

    def get_selected_values(self) -> tuple:
        """
        Returns the values selected by the user.

        :return: Measurement (meas) names that user selects.
        :rtype: tuple

        """
        local_meas = self.local_meas_list.currentItem().text() if self.local_meas_list.currentItem() else None
        remote_meas = self.remote_meas_list.currentItem().text() if self.remote_meas_list.currentItem() else None
        return local_meas, remote_meas


class LayoutSettingsDialog(QDialog):
    """
    It is the layout to show MT field layout settings. Presently it shows the dipole lengths and
    coil numbers only.

    """

    def __init__(self, header, parent=None):
        """
        Constructor

        :param header: Header information of all time series loaded.
        :type header: Dict
        :param parent: Parent
        :type parent: object

        """
        super().__init__(parent)
        self.header = header
        self.setWindowTitle("Layout settings")
        main_layout = QVBoxLayout(self)
        layout = QGridLayout()
        main_layout.addLayout(layout)
        list_of_coils = ['MFS06', 'MFS06e', 'MFS07e']

        if 'ex' in self.header[next(iter(self.header))]:
            self.dipole_north = QLineEdit(str(abs(self.header[next(iter(self.header))]['ex']['x1'][0])))
            self.dipole_south = QLineEdit(str(abs(self.header[next(iter(self.header))]['ex']['x2'][0])))
        if 'ey' in self.header[next(iter(self.header))]:
            self.dipole_east = QLineEdit(str(abs(self.header[next(iter(self.header))]['ey']['y1'][0])))
            self.dipole_west = QLineEdit(str(abs(self.header[next(iter(self.header))]['ey']['y1'][0])))
        #
        if 'hx' in self.header[next(iter(self.header))]:
            self.coil_hx_type = QComboBox()
            self.coil_hx_type.addItems(list_of_coils)
            initial_hx_type = str(self.header[next(iter(self.header))]['hx']['sensor']).replace('\x00', '')
            if initial_hx_type in list_of_coils:
                self.coil_hx_type.setCurrentIndex(self.coil_hx_type.findText(initial_hx_type))
            self.coil_hx = QLineEdit(str(self.header[next(iter(self.header))]['hx']['sensor_no'][0]))
        #
        if 'hy' in self.header[next(iter(self.header))]:
            self.coil_hy_type = QComboBox()
            self.coil_hy_type.addItems(list_of_coils)
            initial_hy_type = str(self.header[next(iter(self.header))]['hy']['sensor']).replace('\x00', '')
            if initial_hy_type in list_of_coils:
                self.coil_hy_type.setCurrentIndex(self.coil_hy_type.findText(initial_hy_type))
            self.coil_hy = QLineEdit(str(self.header[next(iter(self.header))]['hy']['sensor_no'][0]))
        #
        if 'hz' in self.header[next(iter(self.header))]:
            self.coil_hz_type = QComboBox()
            self.coil_hz_type.addItems(list_of_coils)
            initial_hz_type = str(self.header[next(iter(self.header))]['hz']['sensor']).replace('\x00', '')
            if initial_hz_type in list_of_coils:
                self.coil_hz_type.setCurrentIndex(self.coil_hz_type.findText(initial_hz_type))
            self.coil_hz = QLineEdit(str(self.header[next(iter(self.header))]['hz']['sensor_no'][0]))
        #
        if 'rx' in self.header[next(iter(self.header))] and 'ry' in self.header[next(iter(self.header))]:
            #
            self.coil_rx_type = QComboBox()
            self.coil_rx_type.addItems(list_of_coils)
            initial_rx_type = str(self.header[next(iter(self.header))]['rx']['sensor']).replace('\x00', '')
            if initial_rx_type in list_of_coils:
                self.coil_rx_type.setCurrentIndex(self.coil_rx_type.findText(initial_rx_type))
            self.coil_rx = QLineEdit(str(self.header[next(iter(self.header))]['rx']['sensor_no'][0]))
            #
            self.coil_ry_type = QComboBox()
            self.coil_ry_type.addItems(list_of_coils)
            initial_ry_type = str(self.header[next(iter(self.header))]['ry']['sensor']).replace('\x00', '')
            if initial_ry_type in list_of_coils:
                self.coil_ry_type.setCurrentIndex(self.coil_ry_type.findText(initial_ry_type))
            self.coil_ry = QLineEdit(str(self.header[next(iter(self.header))]['ry']['sensor_no'][0]))
        else:
            self.coil_rx = QLineEdit('None')
            self.coil_ry = QLineEdit('None')

        if 'ex' in self.header[next(iter(self.header))]:
            layout.addWidget(QLabel("Dipole N (m):"), 0, 0)
            layout.addWidget(self.dipole_north, 0, 1)
            layout.addWidget(QLabel("Dipole S (m):"), 0, 2)
            layout.addWidget(self.dipole_south, 0, 3)
        if 'ey' in self.header[next(iter(self.header))]:
            layout.addWidget(QLabel("Dipole E (m):"), 1, 0)
            layout.addWidget(self.dipole_east, 1, 1)
            layout.addWidget(QLabel("Dipole W (m):"), 1, 2)
            layout.addWidget(self.dipole_west, 1, 3)
        if 'hx' in self.header[next(iter(self.header))]:
            layout.addWidget(QLabel("Coil type Hx:"), 2, 0)
            layout.addWidget(self.coil_hx_type, 2, 1)
            layout.addWidget(QLabel("Coil # Hx:"), 2, 2)
            layout.addWidget(self.coil_hx, 2, 3)
        if 'hy' in self.header[next(iter(self.header))]:
            layout.addWidget(QLabel("Coil type Hy:"), 3, 0)
            layout.addWidget(self.coil_hy_type, 3, 1)
            layout.addWidget(QLabel("Coil # Hy:"), 3, 2)
            layout.addWidget(self.coil_hy, 3, 3)
        if 'hz' in self.header[next(iter(self.header))]:
            layout.addWidget(QLabel("Coil type Hz:"), 4, 0)
            layout.addWidget(self.coil_hz_type, 4, 1)
            layout.addWidget(QLabel("Coil # Hz:"), 4, 2)
            layout.addWidget(self.coil_hz, 4, 3)

        if 'rx' in self.header[next(iter(self.header))] and 'ry' in self.header[next(iter(self.header))]:
            layout.addWidget(QLabel("Coil type Rx:"), 5, 0)
            layout.addWidget(self.coil_rx_type, 5, 1)

            layout.addWidget(QLabel("Coil # Rx:"), 5, 2)
            layout.addWidget(self.coil_rx, 5, 3)
            #
            layout.addWidget(QLabel("Coil type Ry:"), 6, 0)
            layout.addWidget(self.coil_ry_type, 6, 1)

            layout.addWidget(QLabel("Coil # Ry:"), 6, 2)
            layout.addWidget(self.coil_ry, 6, 3)

        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept_vals)
        buttons.rejected.connect(self.reject)
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(buttons)

        main_layout.addLayout(buttons_layout)
        main_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)

    def accept_vals(self) -> None:
        """
        Update the header with the values from the input fields

        """
        for key in self.header:
            self.header[key]['ex']['x1'][0] = float(self.dipole_north.text())
            self.header[key]['ex']['x2'][0] = float(self.dipole_south.text())
            self.header[key]['ey']['y1'][0] = float(self.dipole_east.text())
            self.header[key]['ey']['y2'][0] = float(self.dipole_west.text())
            self.header[key]['hx']['sensor'] = self.coil_hx_type.currentText()
            self.header[key]['hx']['sensor_no'][0] = int(self.coil_hx.text())
            self.header[key]['hy']['sensor'] = self.coil_hy_type.currentText()
            self.header[key]['hy']['sensor_no'][0] = int(self.coil_hy.text())
            self.header[key]['hz']['sensor'] = self.coil_hz_type.currentText()
            self.header[key]['hz']['sensor_no'][0] = int(self.coil_hz.text())
            if 'rx' in self.header[key] and 'ry' in self.header[key]:
                self.header[key]['rx']['sensor'] = self.coil_rx_type.currentText()
                self.header[key]['rx']['sensor_no'][0] = int(self.coil_rx.text())
                self.header[key]['ry']['sensor'] = self.coil_ry_type.currentText()
                self.header[key]['ry']['sensor_no'][0] = int(self.coil_ry.text())
        super().accept()



"""
First landing window for the Metronix specific operations.

"""

import os
import time

import h5py
import numpy as np
import pandas as pd
import xarray as xr
import yaml
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (QMainWindow, QAction, QFileDialog, QMessageBox,
                             QWidget, QVBoxLayout, QComboBox, QLabel,
                             QPushButton, QHBoxLayout, QRadioButton, QGroupBox,
                             QGridLayout, QLineEdit, QSizePolicy, QDialog,
                             QApplication, QProgressDialog)
from scipy import signal

from sigmt.core import data_selection_tools as dstools
from sigmt.core import perform_data_selection as pds
from sigmt.core import plots
from sigmt.core.band_averaging import BandAveraging
from sigmt.core.robust_estimation import RobustEstimation
from sigmt.gui.about_dialog import AboutDialog
from sigmt.gui.edi_merger import EDIMerger
from sigmt.gui.metronix_dialogs import LayoutSettingsDialog
from sigmt.gui.metronix_dialogs import SelectionDialog
from sigmt.gui.project_related.create_project import ProjectSetupDialog
from sigmt.gui.project_related.edit_project import EditProjectSetupDialog
from sigmt.utils import utils
from sigmt.utils.edi import edi_ops
from sigmt.utils.metronix import metronix_utils

if hasattr(Qt, 'AA_EnableHighDpiScaling'):
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)

if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)


class MainWindow(QMainWindow):
    """
    Definition of the main window for Metronix processing

    """

    def __init__(self):
        """
        Constructor

        """
        super().__init__()
        self.notch_status = None
        self.md_threshold_entry = None
        self.coh_plot_button = None
        self.pd_plot_button = None
        self.about_dialog = None
        self.add_remote_manual_button = None
        self.read_time_series_button = None
        self.decimation_list_dropdown = None
        self.decimate_button = None
        self.new_fs = None
        self.notch_radio_off = None
        self.notch_radio_on = None
        self.fft_length_dropdown = None
        self.parzen_radius_entry = None
        self.save_parameters_button = None
        self.perform_bandavg_button = None
        self.coherency_threshold_edit = None
        self.minimum_data_edit = None
        self.apply_coh_thresh_button = None
        self.clear_coh_thresh_button = None
        self.pd_min_edit = None
        self.pd_max_edit = None
        self.pd_combo_box = None
        self.apply_pd_thresh_button = None
        self.clear_pd_thresh_button = None
        self.plot_impedance_button = None
        self.plot_tipper_button = None
        self.plot_coherency_button = None
        self.save_as_edi_button = None
        self.open_edi_merger_button = None
        self.allsites = None
        self.localsite_path = None
        self.remotesite_path = None
        self.overlapping_meas = None
        self.sfreq_selected = None
        self.processing_df = None
        self.header = None
        self.xml_caldata = None
        self.menubar = None
        self.edi_merger = None
        self.plot_edi_button = None
        self.estimates = None
        self.perform_robust_estimation_button = None
        self.bandavg_dataset = None
        self.procinfo = {}
        self.h5file = None
        self.project_dir = None
        self.project_setup = None
        self.localsite = None
        self.localsite_meas = None
        self.localsite_dropdown = None
        self.remotesite_dropdown = None
        self.sampling_frequency_dropdown = None
        self.verify_layout_button = None
        self.fft_values = ['262144', '131072', '65536', '32768', '16384',
                           '8192', '4096', '2048', '1024', '512', '256']
        #
        self.remotesite = None
        self.remotesite_meas = None
        self.remotesite_manual = None
        #
        self.processing_route = None
        #
        self.interface = 'Metronix'
        self.init_ui()

    def init_ui(self) -> None:
        """
        Initial display user interface

        :return: None
        :rtype: NoneType

        """
        self.setWindowTitle('[No Project Opened] SigMT | A Tool for '
                            'Magnetotelluric Data Processing (Metronix)')
        self.setWindowIcon(QIcon(r'sigmt\images\sigmt.ico'))
        self.setGeometry(100, 100, 700, 500)

        # Creating a menubar
        self.menubar = self.menuBar()

        # Creating File menubar item
        # -------------------------------
        file = self.menubar.addMenu("File")

        create_project_action = QAction("Create a Project", self)
        create_project_action.triggered.connect(self.create_project)
        file.addAction(create_project_action)

        open_project_action = QAction("Open a Project", self)
        open_project_action.triggered.connect(self.open_project)
        file.addAction(open_project_action)

        close_project_action = QAction("Close Project", self)
        close_project_action.triggered.connect(self.close_project)
        file.addAction(close_project_action)

        exit_action = QAction("Exit", self)
        exit_action.triggered.connect(self.exit_all)
        file.addAction(exit_action)

        # Creating Edit menubar item
        # -------------------------------
        edit = self.menubar.addMenu("Edit")

        edit_project_action = QAction("Edit Project Setup", self)
        edit_project_action.triggered.connect(self.edit_project_setup)
        edit.addAction(edit_project_action)

        about_menu = self.menubar.addMenu("About")
        about_action = QAction("About", self)
        about_action.triggered.connect(self.show_about_dialog)
        about_menu.addAction(about_action)

        # Central Widget
        central_widget = QWidget(self)
        self.setCentralWidget(central_widget)

        # Main layout
        main_layout = QVBoxLayout()

        # Subsection 1
        section1 = QGroupBox("Selection of sites and sampling frequency")
        section1_layout = QVBoxLayout()
        section1.setLayout(section1_layout)
        #
        load_sites_widget = QWidget()
        load_sites_layout = QHBoxLayout()
        load_sites_layout.setContentsMargins(0, 0, 0, 0)
        load_sites_widget.setLayout(load_sites_layout)
        load_sites_button = QPushButton("Load Sites")
        load_sites_button.clicked.connect(self.load_sites)
        load_sites_layout.addWidget(load_sites_button)
        load_sites_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        section1_layout.addWidget(load_sites_widget)
        #
        site_selection_widget = QWidget()
        site_selection_layout = QHBoxLayout()
        site_selection_layout.setContentsMargins(0, 0, 0, 0)
        site_selection_widget.setLayout(site_selection_layout)
        site_selection_layout.addWidget(QLabel("Select Local Site:"))
        self.localsite_dropdown = QComboBox()
        self.localsite_dropdown.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        site_selection_layout.addWidget(self.localsite_dropdown)
        site_selection_layout.addWidget(QLabel("Select Remote Site:"))
        self.remotesite_dropdown = QComboBox()
        self.remotesite_dropdown.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        site_selection_layout.addWidget(self.remotesite_dropdown)
        section1_layout.addWidget(site_selection_widget)
        #
        remote_manual_widget = QWidget()
        remote_manual_layout = QHBoxLayout()
        remote_manual_layout.setContentsMargins(0, 0, 0, 0)
        remote_manual_widget.setLayout(remote_manual_layout)
        self.add_remote_manual_button = QPushButton("Set Remote Measurement Manually")
        self.add_remote_manual_button.clicked.connect(self.open_manual_selection_dialog)
        remote_manual_layout.addWidget(self.add_remote_manual_button)
        remote_manual_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        section1_layout.addWidget(remote_manual_widget)
        #
        section1_grid_widget = QWidget()
        section1_grid_layout = QGridLayout()
        section1_grid_layout.setContentsMargins(0, 0, 0, 0)
        section1_grid_widget.setLayout(section1_grid_layout)
        section1_grid_layout.addWidget(QLabel("Select Sampling Frequency"), 0, 0)
        self.sampling_frequency_dropdown = QComboBox()
        section1_grid_layout.addWidget(self.sampling_frequency_dropdown, 0, 1, 1, 40)
        self.read_time_series_button = QPushButton("Read time series")
        self.read_time_series_button.clicked.connect(self.read_ts)
        section1_grid_layout.addWidget(self.read_time_series_button, 1, 0)
        self.verify_layout_button = QPushButton("Verify/Edit layout settings")
        self.verify_layout_button.clicked.connect(self.open_layout_settings)
        self.verify_layout_button.hide()
        section1_grid_layout.addWidget(self.verify_layout_button, 1, 1)
        section1_grid_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        #
        section1_layout.addWidget(section1_grid_widget)
        section1_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)

        # Subsection 2
        section2 = QGroupBox("Decimation")
        section2_layout = QGridLayout()
        section2_layout.addWidget(QLabel("Select Decimation from the list:"), 0, 0)
        self.decimation_list_dropdown = QComboBox()
        self.decimation_list_dropdown.addItems(['4', '8'])
        section2_layout.addWidget(self.decimation_list_dropdown, 0, 1)
        self.decimate_button = QPushButton("Decimate")
        self.decimate_button.clicked.connect(self.decimate)
        section2_layout.addWidget(self.decimate_button, 0, 2)
        self.new_fs = QLabel()
        section2_layout.addWidget(self.new_fs, 0, 3)
        section2.setLayout(section2_layout)
        section2_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)

        # Subsection 3
        section3 = QGroupBox("Parameter Selection")
        section3_layout = QGridLayout()
        section3_layout.addWidget(QLabel("Notch filter:"), 0, 0)
        # ----Notch radio button layout starts----
        self.notch_status = 'off'
        notch_radio_layout = QHBoxLayout()
        self.notch_radio_off = QRadioButton("Off")
        self.notch_radio_on = QRadioButton("On")
        self.notch_radio_off.setChecked(True)
        notch_radio_layout.addWidget(self.notch_radio_off)
        notch_radio_layout.addWidget(self.notch_radio_on)
        #
        self.notch_radio_off.clicked.connect(self.handle_radio_button)
        self.notch_radio_on.clicked.connect(self.handle_radio_button)
        # ----Notch radio button layout ends----
        section3_layout.addLayout(notch_radio_layout, 0, 1)
        section3_layout.addWidget(QLabel("FFT Length"), 1, 0)
        self.fft_length_dropdown = QComboBox()
        section3_layout.addWidget(self.fft_length_dropdown, 1, 1)
        section3_layout.addWidget(QLabel("Parzen window radius"), 1, 2)
        self.parzen_radius_entry = QLineEdit()
        section3_layout.addWidget(self.parzen_radius_entry, 1, 3)
        section3_layout.addWidget(QLabel("Mahalanobis Distance threshold"), 2, 0)
        self.md_threshold_entry = QLineEdit()
        section3_layout.addWidget(self.md_threshold_entry, 2, 1)
        self.save_parameters_button = QPushButton("Save Parameters")
        self.save_parameters_button.clicked.connect(self.save_parameters)
        section3_layout.addWidget(self.save_parameters_button, 3, 0)
        #
        section3.setLayout(section3_layout)
        section3_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)

        # Subsection 4
        section4 = QGroupBox("Band Averaging")
        section4_layout = QGridLayout()
        self.perform_bandavg_button = QPushButton("Perform Band Averaging")
        self.perform_bandavg_button.clicked.connect(self.perform_bandavg)
        section4_layout.addWidget(self.perform_bandavg_button, 0, 0)
        section4.setLayout(section4_layout)
        section4_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)

        # Subsection 5
        section5 = QGroupBox("Data Selection")
        section5_layout = QVBoxLayout()
        # --- Coherency ---
        section5_coh_layout = QHBoxLayout()
        self.coh_plot_button = QPushButton("Check Coherency Plots")
        self.coh_plot_button.clicked.connect(self.plot_coh_all)
        section5_coh_layout.addWidget(self.coh_plot_button)
        section5_coh_layout.addWidget(QLabel("Coherency Threshold"))
        self.coherency_threshold_edit = QLineEdit()
        self.coherency_threshold_edit.setText("0.9")
        section5_coh_layout.addWidget(self.coherency_threshold_edit)
        section5_coh_layout.addWidget(QLabel("Min % Data"))
        self.minimum_data_edit = QLineEdit()
        self.minimum_data_edit.setText("20")
        section5_coh_layout.addWidget(self.minimum_data_edit)
        self.apply_coh_thresh_button = QPushButton("Apply coherency threshold")
        self.apply_coh_thresh_button.clicked.connect(self.apply_coh_thresh)
        section5_coh_layout.addWidget(self.apply_coh_thresh_button)
        self.clear_coh_thresh_button = QPushButton("Clear")
        self.clear_coh_thresh_button.clicked.connect(self.clear_coh_thresh)
        section5_coh_layout.addWidget(self.clear_coh_thresh_button)
        section5_layout.addLayout(section5_coh_layout)
        # --- Polarization direction ---
        section5_pd_layout = QHBoxLayout()
        self.pd_plot_button = QPushButton("Check Polarization Direction (PD) Plots")
        self.pd_plot_button.clicked.connect(self.plot_pd_all)
        section5_pd_layout.addWidget(self.pd_plot_button)
        section5_pd_layout.addWidget(QLabel("PD Min:"))
        self.pd_min_edit = QLineEdit()
        self.pd_min_edit.setText("-10")
        section5_pd_layout.addWidget(self.pd_min_edit)
        section5_pd_layout.addWidget(QLabel("PD Max:"))
        self.pd_max_edit = QLineEdit()
        self.pd_max_edit.setText("10")
        section5_pd_layout.addWidget(self.pd_max_edit)
        section5_pd_layout.addWidget(QLabel("Component:"))
        self.pd_combo_box = QComboBox()
        self.pd_combo_box.addItem("Electric")
        self.pd_combo_box.addItem("Magnetic")
        section5_pd_layout.addWidget(self.pd_combo_box)
        self.apply_pd_thresh_button = QPushButton("Perform PD thresholding")
        self.apply_pd_thresh_button.clicked.connect(self.apply_pd_thresh)
        section5_pd_layout.addWidget(self.apply_pd_thresh_button)
        self.clear_pd_thresh_button = QPushButton("Clear")
        self.clear_pd_thresh_button.clicked.connect(self.clear_pd_thresh)
        section5_pd_layout.addWidget(self.clear_pd_thresh_button)
        section5_layout.addLayout(section5_pd_layout)
        # --------
        section5.setLayout(section5_layout)
        section5_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)

        club_six_seven_eight = QHBoxLayout()

        # Subsection 6
        section6 = QGroupBox("Robust Estimation")
        section6_layout = QHBoxLayout()
        self.perform_robust_estimation_button = QPushButton("Perform Robust Estimation")
        self.perform_robust_estimation_button.clicked.connect(self.perform_robust_estimation)
        section6_layout.addWidget(self.perform_robust_estimation_button)
        section6.setLayout(section6_layout)
        section6_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        club_six_seven_eight.addWidget(section6)

        # Subsection 7
        section7 = QGroupBox("Plots")
        section7_layout = QHBoxLayout()
        self.plot_impedance_button = QPushButton("App. Res and Phase")
        self.plot_impedance_button.clicked.connect(self.plot_impedance)
        section7_layout.addWidget(self.plot_impedance_button)
        self.plot_tipper_button = QPushButton("Tipper")
        section7_layout.addWidget(self.plot_tipper_button)
        self.plot_tipper_button.clicked.connect(self.plot_tipper)

        self.plot_coherency_button = QPushButton("Coherency")
        section7_layout.addWidget(self.plot_coherency_button)
        self.plot_coherency_button.clicked.connect(self.plot_coherency)

        section7.setLayout(section7_layout)
        section7_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        club_six_seven_eight.addWidget(section7)

        # Subsection 8
        section8 = QGroupBox("EDI Operations")
        section8_layout = QHBoxLayout()
        self.save_as_edi_button = QPushButton("Save as EDI")
        section8_layout.addWidget(self.save_as_edi_button)
        self.save_as_edi_button.clicked.connect(self.save_as_edi)
        #
        self.open_edi_merger_button = QPushButton("Open EDI merger")
        section8_layout.addWidget(self.open_edi_merger_button)
        self.open_edi_merger_button.clicked.connect(self.open_edi_merger)
        #
        self.plot_edi_button = QPushButton("Plot an EDI")
        section8_layout.addWidget(self.plot_edi_button)
        self.plot_edi_button.setEnabled(False)
        section8.setLayout(section8_layout)
        section8_layout.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        club_six_seven_eight.addWidget(section8)

        club_six_seven_eight.setAlignment(Qt.AlignTop | Qt.AlignLeft)

        # Adding subsections to the main layout
        main_layout.addWidget(section1)
        main_layout.addWidget(section2)
        main_layout.addWidget(section3)
        main_layout.addWidget(section4)
        main_layout.addWidget(section5)
        main_layout.addLayout(club_six_seven_eight)
        main_layout.setAlignment(Qt.AlignTop)

        # Set main layout to central widget
        central_widget.setLayout(main_layout)

    def create_project(self) -> None:
        """
        It opens the dialog box when the 'create project' option is
        selected from the menu bar.
        A project is mandatory for further processing.

        :return: None
        :rtype: NoneType

        """
        # Prompt user to select project directory
        self.project_dir = QFileDialog.getExistingDirectory(self, "Select Project Directory", "/")
        if self.project_dir:
            # Create the necessary folders
            folders = ['time_series', 'edi', 'project_setup', 'calibration_files']
            for folder in folders:
                folder_path = os.path.join(self.project_dir, folder)
                os.makedirs(folder_path, exist_ok=True)

            # Ask questions and save project setup
            dialog = ProjectSetupDialog(self, interface=self.interface)
            if dialog.exec_():
                self.project_setup = dialog.get_values()

                # Save setup data to YAML file
                setup_file_path = os.path.join(self.project_dir, 'project_setup', 'setup.yaml')
                with open(setup_file_path, 'w') as yaml_file:
                    yaml.dump(self.project_setup, yaml_file)
                QMessageBox.information(self, 'Done',
                                        f'Project created at: {self.project_dir}')
                QMessageBox.information(self, 'Information',
                                        f'Opening new project: {self.project_dir}')
                self.setWindowTitle(self.project_setup[
                                        'project_name'] +
                                    ' - SigMT | '
                                    'A Tool for Magnetotelluric Data Processing (Metronix)')

    def open_project(self) -> None:
        """
        To open an existing project.
        
        :return: None
        :rtype: NoneType

        """
        if self.verify_layout_button is not None:
            self.verify_layout_button.hide()
        self.new_fs.hide()
        self.fft_length_dropdown.clear()
        self.parzen_radius_entry.clear()
        self.md_threshold_entry.clear()
        #
        # Prompt user to select project directory
        self.project_dir = QFileDialog.getExistingDirectory(
            self, "Select Project Directory", "/")
        try:
            self.project_setup = utils.read_yaml_file(os.path.join(
                self.project_dir, 'project_setup', 'setup.yaml'))
            if self.project_setup['interface'] == self.interface:
                self.setWindowTitle(self.project_setup[
                                        'project_name'] +
                                    ' - SigMT | '
                                    'A Tool for Magnetotelluric Data Processing (Metronix)')
            else:
                QMessageBox.warning(self, "Warning", "Not a Metronix Project")
        except:
            QMessageBox.warning(self, "Warning", "Not a valid SigMT Project")

    def close_project(self) -> None:
        """
        Closes the current project and re-opens the MainWindow.

        :return: None
        :rtype: NoneType

        """
        self.close()
        self.__init__()
        self.show()

    def edit_project_setup(self) -> None:
        """
        If user wants to edit the project setup, it can be done using the option in menu bar. This 
        method manages the project setup editing.

        :return: None
        :rtype: NoneType

        """
        if self.project_setup is not None:
            try:
                dialog = EditProjectSetupDialog(
                    self, interface=self.interface, project_setup=self.project_setup)
                if dialog.exec_():
                    self.project_setup = dialog.get_values()

                    # Save setup data to YAML file
                    setup_file_path = os.path.join(self.project_dir, 'project_setup', 'setup.yaml')
                    with open(setup_file_path, 'w') as yaml_file:
                        yaml.dump(self.project_setup, yaml_file)
                    QMessageBox.information(
                        self, 'Done', f'Project setup edited at: {self.project_dir}')
                self.setWindowTitle(self.project_setup[
                                        'project_name'] +
                                    ' - SigMT | '
                                    'A Tool for Magnetotelluric Data Processing (Metronix)')
            except Exception as e:
                QMessageBox.critical(self, "Error", f"An error occurred while editing setup: {e}")
        else:
            QMessageBox.warning(self, "Warning", "Please create and/or open a project!")

    def show_about_dialog(self):
        """
        Opens a dialog box with some information about SigMT and version.

        :return: None
        :rtype: NoneType

        """
        self.about_dialog = AboutDialog()
        self.about_dialog.show()

    def load_sites(self) -> None:
        """
        Method to load the sites in the time_series folder in the project folder.
        It displays all possible sites for local and remote reference processing.

        :return: None
        :rtype: NoneType

        """
        # Resetting some buttons
        self.apply_coh_thresh_button.setText("Apply coherency threshold")
        self.apply_pd_thresh_button.setText("Perform PD thresholding")
        #
        if self.verify_layout_button is not None:
            self.verify_layout_button.hide()
        self.new_fs.hide()
        self.fft_length_dropdown.clear()
        self.parzen_radius_entry.clear()
        self.md_threshold_entry.clear()

        try:
            self.localsite_dropdown.currentIndexChanged.disconnect(self.site_dropdown_changed)
            self.remotesite_dropdown.currentIndexChanged.disconnect(self.site_dropdown_changed)
            self.sampling_frequency_dropdown.clear()
            self.localsite = None
            self.localsite_meas = None
            self.remotesite = None
            self.remotesite_meas = None
            self.remotesite_manual = None
        except:
            pass

        if self.project_dir is not None:
            try:
                self.allsites = metronix_utils.loadsites(self.project_dir)
                if self.localsite_dropdown is not None:
                    self.localsite_dropdown.clear()
                self.localsite_dropdown.addItem("Select---")
                self.localsite_dropdown.addItems(self.allsites)
                # Disable the "Select" option
                self.localsite_dropdown.model().item(0).setEnabled(False)
                self.localsite_dropdown.currentIndexChanged.connect(self.site_dropdown_changed)
                if self.remotesite_dropdown is not None:
                    self.remotesite_dropdown.clear()
                self.remotesite_dropdown.addItem("Select---")
                self.remotesite_dropdown.addItems(self.allsites)
                self.remotesite_dropdown.currentIndexChanged.connect(self.site_dropdown_changed)
                # Disable the "Select" option
                # self.remotesite_dropdown.model().item(0).setEnabled(False)
            except Exception as e:
                QMessageBox.critical(self, "Error", f"An error occurred while loading sites: {e}")
        else:
            QMessageBox.warning(self, "Warning", "Please create and/or open a project!")
            return

    def site_dropdown_changed(self) -> None:
        """
        Gets current text from local site and remote site dropdown menus and finds matching
        meas folders in both. Gets sampling frequencies available in overlapping measurements.
        Displays in the sampling frequency dropdown.
        Two things are checked:
        1. If folder names are same - First and must condition
        2. Folders with same names have same type of measurements? (fs and chopper status)
        If two above conditions are met, the folders are considered as overlapping measurements.
        Added: If folder names are not matching, user will be asked to select remote meas manually.

        :TODO Need to improve strategy later. Works for now!

        :return: None
        :rtype: NoneType

        """
        if self.verify_layout_button is not None:
            self.verify_layout_button.hide()
        self.new_fs.hide()
        self.fft_length_dropdown.clear()
        self.parzen_radius_entry.clear()
        self.md_threshold_entry.clear()
        #
        self.remotesite_manual = None
        self.localsite = self.localsite_dropdown.currentText()
        if self.remotesite_dropdown.currentText() == "Select---":
            self.remotesite = None
            self.remotesite_meas = None
        else:
            self.remotesite = self.remotesite_dropdown.currentText()
        # List meas folders in local site
        self.localsite_path = os.path.join(self.project_dir, 'time_series', self.localsite)
        self.localsite_meas = metronix_utils.list_meas_folders(self.localsite_path)
        # List meas folders in remote site
        if self.remotesite is not None:
            self.remotesite_path = os.path.join(self.project_dir, 'time_series', self.remotesite)
            self.remotesite_meas = metronix_utils.list_meas_folders(self.remotesite_path)
        # Finding overlapping measurements
        if self.remotesite is not None:
            overlapping_meas = sorted(
                list(set(self.localsite_meas).intersection(set(self.remotesite_meas))))
            # If no overlapping found with remote, return to single site mode
            if not overlapping_meas:
                QMessageBox.warning(self, 'Warning',
                                    "No overlapping measurements found! You need to "
                                    "set remote measurement manually "
                                    "by clicking 'Set Remote Measurement Manually' button below. \n\n"
                                    "Else, please proceed with single site processing.")
                # self.remotesite_dropdown.blockSignals(True)
                # self.remotesite_dropdown.setCurrentIndex(0)
                # self.remotesite_dropdown.blockSignals(False)
                self.remotesite_manual = self.remotesite
                self.remotesite = None
                overlapping_meas = self.localsite_meas
        elif self.remotesite is None:
            overlapping_meas = self.localsite_meas
        # Operations on local site
        local_meas_paths = [os.path.join(self.project_dir, 'time_series', self.localsite, meas) for
                            meas in
                            overlapping_meas]
        [sampfreq, chopper_value] = metronix_utils.get_sampling_frequency_from_xml(local_meas_paths)
        localsite_meas_overlap = set(zip(overlapping_meas, sampfreq, chopper_value))
        # Operations on remote site
        if self.remotesite is not None:
            remote_meas_paths = [
                os.path.join(self.project_dir, 'time_series', self.remotesite, meas) for meas in
                overlapping_meas]
            [sampfreq, chopper_value] = metronix_utils.get_sampling_frequency_from_xml(
                remote_meas_paths)
            remotesite_meas_overlap = set(zip(overlapping_meas, sampfreq, chopper_value))
        if self.remotesite is not None:
            self.overlapping_meas = list(
                localsite_meas_overlap.intersection(remotesite_meas_overlap))
        elif self.remotesite is None:
            self.overlapping_meas = list(localsite_meas_overlap)
        # Creating processing route
        # It is a dataframe containing measurement details
        if self.remotesite is not None:
            processing_route = {
                'local': [meas[0] for meas in self.overlapping_meas],
                'remote': [meas[0] for meas in self.overlapping_meas],
                'sampling_frequency': [meas[1] for meas in self.overlapping_meas],
                'chopper_status': [meas[2] for meas in self.overlapping_meas],
                'sampling_chopper': [str(meas[1]) + ' Hz - ' + meas[2] for meas in
                                     self.overlapping_meas]
            }
        elif self.remotesite is None:
            processing_route = {
                'local': [meas[0] for meas in self.overlapping_meas],
                'remote': [None for meas in self.overlapping_meas],
                'sampling_frequency': [meas[1] for meas in self.overlapping_meas],
                'chopper_status': [meas[2] for meas in self.overlapping_meas],
                'sampling_chopper': [str(meas[1]) + ' Hz - ' + meas[2] for meas in
                                     self.overlapping_meas]
            }
        if self.remotesite_manual is not None:
            processing_route = {
                'local': [],
                'remote': [],
                'sampling_frequency': [],
                'chopper_status': [],
                'sampling_chopper': []
            }
        self.processing_route = pd.DataFrame(processing_route)
        unique_samp_values = self.processing_route['sampling_chopper'].unique().tolist()
        # Displaying in sampling frequency dropdown
        self.sampling_frequency_dropdown.clear()  # Clear existing dropdown
        self.sampling_frequency_dropdown.addItems(unique_samp_values)  # Add new values

    def open_manual_selection_dialog(self) -> None:
        """
        Opens a dialog window for manual selection of remote measurements.
        This method creates a dialog window for manually selecting remote measurements.

        :return: None
        :rtype: NoneType

        """
        try:
            dialog = SelectionDialog(local_meas=self.localsite_meas,
                                     remote_meas=self.remotesite_meas, parent=self)
            if dialog.exec_() == QDialog.Accepted:
                [local, remote] = dialog.get_selected_values()
                if self.remotesite is None:
                    self.remotesite = self.remotesite_manual
                local_meas_path = [
                    os.path.join(self.project_dir, 'time_series', self.localsite, local)]
                [l_sampfreq, l_chopper_value] = metronix_utils.get_sampling_frequency_from_xml(
                    local_meas_path)
                remote_meas_path = [
                    os.path.join(self.project_dir, 'time_series', self.remotesite, remote)]
                [r_sampfreq, r_chopper_value] = metronix_utils.get_sampling_frequency_from_xml(
                    remote_meas_path)
                if l_sampfreq != r_sampfreq:
                    QMessageBox.critical(self, "Error",
                                         f"Sampling frequencies are not matching. Cannot proceed.")
                    self.remotesite = None
                    self.remotesite_dropdown.setCurrentIndex(0)
                    return
                if l_chopper_value != r_chopper_value:
                    QMessageBox.critical(self, "Error",
                                         f"Chopper status not matching. Cannot proceed. Choose another match.")
                    self.remotesite = None
                    self.remotesite_dropdown.setCurrentIndex(0)
                    return
                new_row = {
                    'local': local,
                    'remote': remote,
                    'sampling_frequency': r_sampfreq[0],
                    'chopper_status': r_chopper_value[0],
                    'sampling_chopper': str(r_sampfreq[0]) + ' Hz - ' + r_chopper_value[0]
                }
                # Append the new row to the DataFrame
                self.processing_route = self.processing_route._append(new_row, ignore_index=True)
                unique_samp_values = self.processing_route['sampling_chopper'].unique().tolist()
                self.sampling_frequency_dropdown.clear()
                self.sampling_frequency_dropdown.addItems(unique_samp_values)
            self.processing_route.drop_duplicates(inplace=True)
        except:
            QMessageBox.critical(self, "Error",
                                 f"Select local and remote site first! It works only if remote site is selected.")

    # noinspection PyTypeChecker
    def read_ts(self) -> None:
        """
        Reads time series based on the self.processing_route.
        processing_route is a DataFrame containing ['local', 'remote', 'sampling_frequency', 'chopper_status',
        'sampling_chopper'] for all measurements.
        processing_df contains the values for selected sampling frequency & chopper status combo.

        :return: None
        :rtype: NoneType

        """
        # Reseting some buttons
        self.apply_coh_thresh_button.setText("Apply coherency threshold")
        self.apply_pd_thresh_button.setText("Perform PD thresholding")
        #
        self.new_fs.hide()
        if self.processing_route is not None:
            self.sfreq_selected = self.sampling_frequency_dropdown.currentText()
            self.processing_df = self.processing_route[
                self.processing_route['sampling_chopper'] == self.sfreq_selected]
        else:
            QMessageBox.warning(self, 'Warning', "Please choose sampling frequency.")
            return
        if self.localsite is None:
            QMessageBox.warning(self, 'Warning', "Please choose local and/or remote site.")
            return

        # Create a progress dialog
        progress_dialog = QProgressDialog("Reading time series...", None, 0,
                                          len(self.processing_df.index), self)
        progress_dialog.setWindowModality(Qt.WindowModal)
        progress_dialog.setWindowTitle("Please wait")
        progress_dialog.show()

        # This is important to show the progress bar and GUI not to freeze
        qapp_instance = QApplication.instance()
        qapp_instance.processEvents()

        # If single site processing or remote site processing?
        if self.remotesite is None:
            local_paths = [os.path.join(self.project_dir, 'time_series', self.localsite, meas) for
                           meas in
                           self.processing_df['local']]
            num = 0
            self.header = {}
            self.xml_caldata = {}
            self.h5file = os.path.join(self.project_dir, 'db.h5')
            if os.path.exists(self.h5file):
                os.remove(self.h5file)
            # Creates a h5 file in project directory in write mode
            with h5py.File(self.h5file, 'w') as f:
                for local_path in local_paths:
                    self.header[f'ts_{num}'], ts_dict = metronix_utils.read_ts(local_path,
                                                                               self.project_setup)
                    self.xml_caldata[f'ts_{num}'] = metronix_utils.read_calibration_from_xml(
                        local_path)
                    ts = f.create_group(f'ts_{num}')
                    for key in ts_dict.keys():
                        ts.create_dataset(key, data=ts_dict[key].values)
                        self.header[f'ts_{num}'][key]['time_coord'] = ts_dict[key].time
                    del ts_dict
                    num += 1
                    progress_dialog.setValue(num)
        elif self.remotesite is not None:
            local_paths = [os.path.join(self.project_dir, 'time_series', self.localsite, meas) for
                           meas in
                           self.processing_df['local']]
            remote_paths = [os.path.join(self.project_dir, 'time_series', self.remotesite, meas) for
                            meas in
                            self.processing_df['remote']]
            num = 0
            self.header = {}
            self.xml_caldata = {}
            self.h5file = os.path.join(self.project_dir, 'db.h5')
            if os.path.exists(self.h5file):
                os.remove(self.h5file)
            # Creates a h5 file in project directory in write mode
            with h5py.File(self.h5file, 'w') as f:
                for local_path, remote_path in zip(local_paths, remote_paths):
                    self.header[f'ts_{num}'], ts_dict = metronix_utils.read_ts(local_path,
                                                                               self.project_setup)
                    self.xml_caldata[f'ts_{num}'] = metronix_utils.read_calibration_from_xml(
                        local_path)
                    ts = f.create_group(f'ts_{num}')
                    for key in ts_dict.keys():
                        ts.create_dataset(key, data=ts_dict[key].values)
                        self.header[f'ts_{num}'][key]['time_coord'] = ts_dict[key].time
                        # self.header[f'ts_{num}'][key]['start_time'] = ts_dict[key].time.min()
                        # self.header[f'ts_{num}'][key]['end_time'] = ts_dict[key].time.max()
                    header_r, ts_r = metronix_utils.read_ts(remote_path, self.project_setup)
                    xml_caldata_r = metronix_utils.read_calibration_from_xml(remote_path)
                    self.xml_caldata[f'ts_{num}'].update(xml_caldata_r)
                    if 'hx' in header_r:
                        mask = ts_r['hx']['time'].isin(ts_dict['hx']['time'])
                        ts_r['hx'] = ts_r['hx'].sel(time=mask)
                        header_r['hx']['nsamples'] = len(ts_r['hx'])
                        self.header[f'ts_{num}']['rx'] = header_r['hx']
                        ts.create_dataset('rx', data=ts_r['hx'].values)
                        self.header[f'ts_{num}']['rx']['time_coord'] = ts_r['hx'].time
                        # self.header[f'ts_{num}']['Rx']['start_time'] = ts_r['Hx'].time.min()
                        # self.header[f'ts_{num}']['Rx']['end_time'] = ts_r['Hx'].time.max()
                    if 'hy' in header_r:
                        mask = ts_r['hy']['time'].isin(ts_dict['hy']['time'])
                        ts_r['hy'] = ts_r['hy'].sel(time=mask)
                        header_r['hy']['nsamples'] = len(ts_r['hy'])
                        self.header[f'ts_{num}']['ry'] = header_r['hy']
                        ts.create_dataset('ry', data=ts_r['hy'].values)
                        self.header[f'ts_{num}']['ry']['time_coord'] = ts_r['hy'].time
                        # self.header[f'ts_{num}']['Ry']['start_time'] = ts_r['Hy'].time.min()
                        # self.header[f'ts_{num}']['Ry']['end_time'] = ts_r['Hy'].time.max()
                        del header_r, ts_r
                    else:
                        QMessageBox.information(self, "Information",
                                                f"Hx and/or Hy are not found at {remote_path}. "
                                                f"Skipping to next measurement.")
                        return
                    del ts_dict
                    num += 1
                    progress_dialog.setValue(num)
        qapp_instance.processEvents()
        progress_dialog.close()
        self.procinfo['fs'] = self.processing_df['sampling_frequency'].iloc[0]
        QMessageBox.information(self, "Information", "Time series reading completed!\n\n"
                                                     "Summary\n"
                                                     "----------\n"
                                                     f"No. of measurements loaded: {len(self.processing_df)}\n"
                                                     f"Sampling frequency: {self.procinfo['fs']} Hz")
        self.procinfo['nsamples_mostly'] = utils.get_nsamples(self.header)
        fftlength = utils.get_fftlength(self.procinfo['nsamples_mostly'])
        # Updating FFT length dropdown
        self.fft_length_dropdown.addItems(self.fft_values)
        self.fft_length_dropdown.setCurrentIndex(self.fft_length_dropdown.findText(str(fftlength)))
        # Updating parzen window radius
        parzen_radius = utils.get_parzen(self.procinfo['fs'])
        self.parzen_radius_entry.setText(str(parzen_radius))
        self.verify_layout_button.show()
        # Updating MD thresholds
        self.md_threshold_entry.setText(str(1.5))

    def open_layout_settings(self) -> None:
        """
        Opens layout settings dialog

        :return: None
        :rtype: NoneType

        """
        try:
            dialog = LayoutSettingsDialog(header=self.header, parent=self)
            if dialog.exec_() == QDialog.Accepted:
                self.header = dialog.header
        except Exception as e:
            QMessageBox.critical(self, "Error", f"An error occurred while layout settings: {e}")

    def decimate(self) -> None:
        """
        Decimate the time series.

        :return: None
        :rtype: NoneType

        """
        # Reseting some buttons
        self.apply_coh_thresh_button.setText("Apply coherency threshold")
        self.apply_pd_thresh_button.setText("Perform PD thresholding")
        #
        if self.localsite is None:
            QMessageBox.warning(self, 'Warning', "Please choose local and/or remote site.")
            return
        try:
            progress_dialog = QProgressDialog("Performing Decimation...", None, 0, 1, self)
            progress_dialog.setWindowModality(Qt.WindowModal)
            progress_dialog.setWindowTitle("Please wait")
            progress_dialog.show()
            # This is important to show the progress bar and GUI not to freeze
            qapp_instance = QApplication.instance()
            qapp_instance.processEvents()
            decimation_factor = int(self.decimation_list_dropdown.currentText())
            with h5py.File(self.h5file, 'r+') as f:
                for ts in f.keys():
                    for channel in f[ts]:
                        decimated_data = signal.decimate(f[ts][channel][:], decimation_factor,
                                                         n=None, ftype='iir')
                        del f[ts][channel]
                        ts_group = f[ts]
                        ts_group.create_dataset(channel, data=decimated_data)
                        self.header[ts][channel]['sfreq'] = [
                            self.header[ts][channel]['sfreq'][0] / decimation_factor]
                        self.header[ts][channel]['nsamples'] = len(decimated_data)
                        self.header[ts][channel]['time_coord'] = self.header[ts][channel][
                                                                     'time_coord'][
                                                                 ::decimation_factor]
            self.procinfo['fs'] = self.procinfo['fs'] / decimation_factor
            self.new_fs.setText(f"Now sampling frequency is {self.procinfo['fs']} Hz")
            self.new_fs.show()
            self.procinfo['nsamples_mostly'] = utils.get_nsamples(self.header)
            fftlength = utils.get_fftlength(self.procinfo['nsamples_mostly'])
            # Updating FFT length dropdown
            self.fft_length_dropdown.setCurrentIndex(
                self.fft_length_dropdown.findText(str(fftlength)))
            # Updating parzen window radius
            parzen_radius = utils.get_parzen(self.procinfo['fs'])
            self.parzen_radius_entry.setText(str(parzen_radius))
            #
            progress_dialog.setValue(1)
            qapp_instance.processEvents()
            QMessageBox.information(self, "Done", "Decimation done!")

        except Exception as e:
            QMessageBox.critical(self, "Error", f"An error occurred while decimation: {e}")

    def handle_radio_button(self) -> None:
        """
        Handles the state change of the radio buttons and updates the notch filter status.

        :return: None
        :rtype: NoneType

        """
        if self.notch_radio_off.isChecked():
            self.notch_status = 'off'
        elif self.notch_radio_on.isChecked():
            self.notch_status = 'on'

    def save_parameters(self) -> None:
        """
        Saves the parameters to the self.procinfo dictionary.

        :return: None
        :rtype: NoneType

        """
        # Reseting some buttons
        self.apply_coh_thresh_button.setText("Apply coherency threshold")
        self.apply_pd_thresh_button.setText("Perform PD thresholding")
        #
        try:
            if self.header is not None:
                self.procinfo['localsite'] = self.localsite
                self.procinfo['remotesite'] = self.remotesite
                self.procinfo['notch'] = self.notch_status
                self.procinfo['processing_mode'] = self.project_setup['processing_mode']
                self.procinfo['fft_length'] = int(self.fft_length_dropdown.currentText())
                self.procinfo['parzen_radius'] = float(self.parzen_radius_entry.text())
                self.procinfo['md_thresh'] = float(self.md_threshold_entry.text())
                self.procinfo['notch_frequency'] = float(self.project_setup['notch_frequency'])
                self.procinfo['preferred_cal_file'] = self.project_setup['preferred_cal_file']
                self.procinfo['frequencies_per_decade'] = int(
                    self.project_setup['frequencies_per_decade'])
                first_header = next(iter(next(iter(self.header.values())).values()))
                #
                self.procinfo['lat'] = first_header['lat'][0] / 1000 / 60 / 60
                self.procinfo['lon'] = first_header['lon'][0] / 1000 / 60 / 60
                self.procinfo['elev'] = first_header['elev'][0] / 100
                #
                self.procinfo['start_time'] = first_header['start'][0]
                QMessageBox.information(self, "Done", f"Parameters are saved.")
            else:
                QMessageBox.critical(self, "Error", f"Please read time series first!!")
        except Exception as e:
            QMessageBox.critical(self, "Error", f"An error occurred: {str(e)}."
                                                "Ensure time series is read!!!")

    def perform_bandavg(self) -> None:
        """
        Set up the band averaging.

        :return: None
        :rtype: NoneType

        """
        # Resetting some buttons
        self.apply_coh_thresh_button.setText("Apply coherency threshold")
        self.apply_pd_thresh_button.setText("Perform PD thresholding")
        #
        if 'localsite' in self.procinfo and self.procinfo['localsite'] == self.localsite:
            datasets = []
            #
            progress_dialog = QProgressDialog("Performing band averaging...", None, 0,
                                              len(self.header), self)
            progress_dialog.setWindowModality(Qt.WindowModal)
            progress_dialog.setWindowTitle("Please wait")
            progress_dialog.show()
            # This is important to show the progress bar and GUI not to freeze
            qapp_instance = QApplication.instance()
            qapp_instance.processEvents()
            #
            num = 0
            bandavg_time = time.time()
            for ts in self.header:

                # Preparing inputs for band averaging
                # TODO: Replace this with a better strategy later
                if self.procinfo['notch'] == 'on':
                    notch_filter_apply = True
                else:
                    notch_filter_apply = False

                # TODO: Replace this with a better strategy later
                if self.procinfo['processing_mode'] == 'MT + Tipper':
                    process_mt = True
                    process_tipper = True
                elif self.procinfo['processing_mode'] == 'MT Only':
                    process_mt = True
                    process_tipper = False
                else:
                    process_mt = None
                    process_tipper = None

                # TODO: Replace this with a better strategy later
                if self.procinfo['remotesite'] is not None:
                    remote_reference = True
                else:
                    remote_reference = False

                # TODO: Replace this with a better strategy later
                calibration_data_electric = {}
                if 'ex' in self.header[ts]:
                    calibration_data_electric['ex'] = {}
                    calibration_data_electric['ex']['x1'] = self.header[ts]['ex']['x1'][0]
                    calibration_data_electric['ex']['x2'] = self.header[ts]['ex']['x2'][0]

                if 'ey' in self.header[ts]:
                    calibration_data_electric['ey'] = {}
                    calibration_data_electric['ey']['y1'] = self.header[ts]['ey']['y1'][0]
                    calibration_data_electric['ey']['y2'] = self.header[ts]['ey']['y2'][0]

                # TODO: Replace this with a better strategy later
                calibration_data_magnetic = {'instrument': 'metronix'}
                possible_magnetic_channels = ['hx', 'hy', 'hz', 'rx', 'ry']
                available_magnetic_channels = [element for element in possible_magnetic_channels if
                                               element in list(self.header[ts].keys())]
                for magnetic_channel in available_magnetic_channels:
                    calibration_data_magnetic[magnetic_channel] = {}
                    calibration_data_magnetic[magnetic_channel]['sensor_type'] = self.header[ts][magnetic_channel]['sensor']
                    calibration_data_magnetic[magnetic_channel]['sensor_serial_number'] = \
                    self.header[ts][magnetic_channel]['sensor_no'][0]
                    calibration_data_magnetic[magnetic_channel]['chopper_status'] = \
                    self.header[ts][magnetic_channel]['bychopper'][0]
                    calibration_data_magnetic[magnetic_channel]['calibration_data'] = self.xml_caldata[ts]

                # Get the bandavg object
                bandavg = BandAveraging(time_series=metronix_utils.prepare_ts_from_h5(self.h5file, ts),
                                        sampling_frequency=self.procinfo['fs'], overlap=50,
                                        calibrate_electric=True, calibrate_magnetic=True,
                                        calibration_data_electric=calibration_data_electric,
                                        calibration_data_magnetic=calibration_data_magnetic,
                                        fft_length=self.procinfo['fft_length'],
                                        parzen_window_radius=self.procinfo['parzen_radius'],
                                        frequencies_per_decade=self.procinfo['frequencies_per_decade'],
                                        apply_notch_filter=notch_filter_apply,
                                        notch_frequency=self.procinfo['notch_frequency'],
                                        process_mt=process_mt, process_tipper=process_tipper,
                                        remote_reference=remote_reference
                                        )
                datasets.append(bandavg.band_averaged_dataset)  # appends xarray dataset (for a run)
                num += 1
                progress_dialog.setValue(num)
            self.bandavg_dataset = xr.concat(datasets, dim='time_window').assign_coords(
                time_window=np.arange(len(xr.concat(datasets, dim='time_window').time_window)))
            self.dof = bandavg.dof
            self.avgf = bandavg.avgf
            self.bandavg_dataset['dof'] = xr.DataArray(
                self.dof,
                coords={'frequency': self.bandavg_dataset.coords['frequency']},
                dims='frequency'
            )
            print(f'Time taken for band averaging: ' + str(time.time() - bandavg_time))
            # Calculating data selection parameters
            time_dataselection = time.time()
            self.bandavg_dataset['coh_ex'] = (
                ('time_window', 'frequency'), dstools.cohex(self.bandavg_dataset))
            self.bandavg_dataset['coh_ey'] = (
                ('time_window', 'frequency'), dstools.cohey(self.bandavg_dataset))
            if not self.project_setup['processing_mode'] == "MT Only":
                self.bandavg_dataset['coh_hz'] = (
                    ('time_window', 'frequency'), dstools.cohhz(self.bandavg_dataset))
            self.bandavg_dataset['alpha_h'], self.bandavg_dataset[
                'alpha_e'] = dstools.pdvalues(
                self.bandavg_dataset)
            print(f'Time taken for data selection tool: ' + str(time.time() - time_dataselection))
            qapp_instance.processEvents()
            QMessageBox.information(self, "Done", f"Band averaging done!")
        else:
            QMessageBox.critical(self, "Error", f"Please read time series again!!")

    def plot_coh_all(self) -> None:
        """
        Plot coherency plots for all frequencies

        :return: None
        :rtype: NoneType

        """
        plots.plot_coherencies_all(self.bandavg_dataset)

    def plot_pd_all(self) -> None:
        """
        Plot polarization direction plots for all frequencies

        """
        plots.plot_pd_all(self.bandavg_dataset)

    def apply_coh_thresh(self) -> None:
        """
        Apply coherency threshold.

        :return: None
        :rtype: NoneType

        """
        if self.project_setup:
            output_channels = None
            coh_thresh = float(self.coherency_threshold_edit.text())
            min_percent = float(self.minimum_data_edit.text())
            if self.project_setup['processing_mode'] == "MT + Tipper":
                output_channels = ['ex', 'ey', 'hz']
            elif self.project_setup['processing_mode'] == "MT Only":
                output_channels = ['ex', 'ey']
            self.bandavg_dataset = pds.perform_coh_thresh(self.bandavg_dataset, coh_thresh,
                                                          min_percent,
                                                          output_channels)

            self.apply_coh_thresh_button.setText("Coherency threshold APPLIED!")

    def clear_coh_thresh(self) -> None:
        """
        Clear coherency threshold value.

        :return: None
        :rtype: NoneType

        """
        if self.bandavg_dataset:
            self.bandavg_dataset = pds.clear_coh_thresh(self.bandavg_dataset)
            self.apply_coh_thresh_button.setText("Apply coherency threshold")

    def apply_pd_thresh(self) -> None:
        """
        Apply polarization direction threshold.

        :return: None
        :rtype: NoneType

        """
        if self.bandavg_dataset:
            if not all(item in self.bandavg_dataset for item in ['ex', 'ey', 'hx', 'hy']):
                QMessageBox.warning(self, "Warning", "One of ['Ex', 'Ey', 'Hx', 'Hy'] is missing.")
            else:
                component = self.pd_combo_box.currentText()
                pd_min = float(self.pd_min_edit.text())
                pd_max = float(self.pd_max_edit.text())
                self.bandavg_dataset = pds.perform_pd_selection(self.bandavg_dataset, component,
                                                                pd_min, pd_max)
                self.apply_pd_thresh_button.setText("PD threshold APPLIED!")

    def clear_pd_thresh(self) -> None:
        """
        Clear polarization direction threshold.

        :return: None
        :rtype: NoneType

        """
        if self.bandavg_dataset:
            self.bandavg_dataset = pds.clear_pd_selection(self.bandavg_dataset)
            self.apply_pd_thresh_button.setText("Perform PD thresholding")

    def perform_robust_estimation(self) -> None:
        """
        Performs robust estimation with UI handling.

        :return: None
        :rtype: NoneType

        """
        if self.bandavg_dataset:
            progress_dialog = QProgressDialog("Performing Robust Estimation...", None, 0, 1, self)
            progress_dialog.setWindowModality(Qt.WindowModal)
            progress_dialog.setWindowTitle("Please wait")
            progress_dialog.show()
            # This is important to show the progress bar and GUI not to freeze
            qapp_instance = QApplication.instance()
            qapp_instance.processEvents()

            # Perform the robust estimation
            estimation_instance = RobustEstimation(self.procinfo, self.bandavg_dataset)
            self.estimates = estimation_instance.estimates

            progress_dialog.setValue(1)
            qapp_instance.processEvents()
            QMessageBox.information(self, "Done", "Robust estimation done!")

    def plot_impedance(self) -> None:
        """
        Plot apparent resitivity and phase.

        :return: None
        :rtype: NoneType

        """
        plots.plot_mt_app_res(self.estimates, self.procinfo)

    def plot_tipper(self) -> None:
        """
        Plot tipper.

        :return: None
        :rtype: NoneType

        """
        if self.project_setup:
            if not self.project_setup['processing_mode'] == "MT Only":
                plots.plot_tipper(self.estimates, self.procinfo)

    def plot_coherency(self) -> None:
        """
        Plot coherency.

        :return: None
        :rtype: NoneType

        """
        plots.plot_coherency(self.estimates, self.procinfo)

    def save_as_edi(self) -> None:
        """
        Save as EDI.

        :return: None
        :rtype: NoneType

        """
        edi_ops.save_edi(self.estimates, self.procinfo, self.project_setup)

    def open_edi_merger(self) -> None:
        """
        Open EDI merger.

        :return: None
        :rtype: NoneType

        """
        self.edi_merger = EDIMerger()
        self.edi_merger.show()

    def exit_all(self) -> None:
        """
        Close the application.

        This method is used to exit the application by closing the main window.
        It calls the self.close() method, which typically triggers the
        application to terminate.

        :return: None
        :rtype: NoneType

        """
        self.close()



"""
Design of the GUI for the editing of project settings.
"""
from PyQt5.QtWidgets import QLabel, QDialog, QLineEdit, QVBoxLayout, QDialogButtonBox, QComboBox


class EditProjectSetupDialog(QDialog):
    """
    GUI to edit project setup
    """

    def __init__(self, parent=None, interface=None, project_setup=None):
        super().__init__(parent)
        self.interface = interface
        self.project_setup = project_setup
        self.setWindowTitle("Edit Project Setup")
        layout = QVBoxLayout()

        self.project_name = QLineEdit(self.project_setup['project_name'])
        self.country = QLineEdit(self.project_setup['country'])
        self.acquired_by_company = QLineEdit(self.project_setup['acquired_by'])
        #
        self.processing_mode = QComboBox()
        self.processing_mode.addItem('MT + Tipper')
        self.processing_mode.addItem('MT Only')
        # TODO: Add this in future
        # self.processing_mode.addItem('Tipper Only')
        self.processing_mode.setCurrentText(self.project_setup['processing_mode'])
        #
        self.frequencies_per_decade = QLineEdit(self.project_setup['frequencies_per_decade'])
        self.notch_frequency = QLineEdit(self.project_setup['notch_frequency'])
        self.preferred_cal_file = QComboBox()
        self.preferred_cal_file.addItem('xml')
        # TODO: Add this in future
        # self.preferred_cal_file.addItem('metronix txt')
        self.preferred_cal_file.setCurrentText(self.project_setup['preferred_cal_file'])

        layout.addWidget(QLabel("Project Name:"))
        layout.addWidget(self.project_name)

        layout.addWidget(QLabel("Country:"))
        layout.addWidget(self.country)

        layout.addWidget(QLabel("Acquired by:"))
        layout.addWidget(self.acquired_by_company)

        layout.addWidget(QLabel("Processing mode:"))
        layout.addWidget(self.processing_mode)

        layout.addWidget(QLabel("Frequencies per Decade:"))
        layout.addWidget(self.frequencies_per_decade)

        layout.addWidget(QLabel("Notch Frequency:"))
        layout.addWidget(self.notch_frequency)

        layout.addWidget(QLabel("Preferred calibration file:"))
        layout.addWidget(self.preferred_cal_file)

        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

        self.setLayout(layout)

    def get_values(self) -> dict:
        """
        Returns dictionary with values

        :return: Dictionary with values
        :rtype: dict
        """
        return {
            "project_name": self.project_name.text(),
            "country": self.country.text(),
            "acquired_by": self.acquired_by_company.text(),
            "processing_mode": self.processing_mode.currentText(),
            "frequencies_per_decade": self.frequencies_per_decade.text(),
            "notch_frequency": self.notch_frequency.text(),
            "interface": self.interface,
            "preferred_cal_file": self.preferred_cal_file.currentText()
        }



"""
Class for the metronix coil calibration
"""
import numpy as np


class MetronixCalibration:
    """
    Class to perform Metronix specific calibration.
    """

    def __init__(self, xfft, fft_freqs, sensor_type, stat, calibration_data):
        """
        Constructor

        :param xfft: FFT values for the time series shape: (fft_length, nof_windows)
        :type xfft: np.ndarray
        :param fft_freqs: Frequencies for FFT values, 1D array.
        :type fft_freqs: np.ndarray
        :param sensor_type: Type of sensor, Eg: MFS06e
        :type sensor_type: str
        :param stat: Chopper status. Either 'chopper_on' or 'chopper_off'
        :type stat: str
        :param calibration_data: Calibration data as array. shape: (length(43), 3).
        :type calibration_data: np.ndarray
        """
        self.calibrated_data = None
        self.xfft = xfft
        self.fft_freqs = fft_freqs
        self.sensor_type = sensor_type
        self.chopper_status = stat
        self.calibration_data = calibration_data
        self.perform_calibration()

    def perform_calibration(self):
        """
        Streamline it based on the sensor type.
        """
        if self.sensor_type == 'MFS06' or 'MFS06e':
            self.mfs06e()
        elif self.sensor_type == 'MFS07e':
            self.mfs07e()

    def mfs06e(self):
        """
        Streamline calibration for mfs06e
        """
        magnitude = self.calibration_data[:, 0] * self.calibration_data[:, 1]
        phase = np.radians(self.calibration_data[:, 2])
        calt = (magnitude * np.cos(phase) + (1j * magnitude * np.sin(phase))) * 1000
        # If Chopper is Off
        if self.chopper_status == 'chopper_off':
            cal_all_band = np.interp(self.fft_freqs, self.calibration_data[:, 0], calt)
            self.calibrated_data = self.xfft / cal_all_band[:, np.newaxis]
        # If Chopper is On
        if self.chopper_status == 'chopper_on':
            minfindx = np.where(self.fft_freqs < 0.1)[0]
            cal_all_band = np.interp(self.fft_freqs[np.max(minfindx) + 1:np.shape(self.fft_freqs)[0]],
                                     self.calibration_data[:, 0], calt)
            thmag = np.zeros(np.shape(minfindx), )
            thmag[:, ] = 0.2 * self.fft_freqs[0:np.max(minfindx) + 1]
            thph = np.arctan2(4.0, self.fft_freqs[0:np.max(minfindx) + 1])
            th_band = (thmag * np.cos(thph) + (1j * thmag * np.sin(thph))) * 1000
            cal_all_band = np.concatenate((th_band, cal_all_band))
            cal_all_band[0] = 1 + 1j  # To avoid division of zero
            self.calibrated_data = self.xfft / cal_all_band[:, np.newaxis]

    def mfs07e(self):
        """
        Streamline calibration for mfs07e
        """
        magnitude = self.calibration_data[:, 0] * self.calibration_data[:, 1]
        phase = np.radians(self.calibration_data[:, 2])
        calt = (magnitude * np.cos(phase) + (1j * magnitude * np.sin(phase))) * 1000
        # If Chopper is Off
        if self.chopper_status == 'chopper_off':
            cal_all_band = np.interp(self.fft_freqs, self.calibration_data[:, 0], calt)
            self.calibrated_data = self.xfft / cal_all_band[:, np.newaxis]
        # If Chopper is On
        if self.chopper_status == 'chopper_on':
            minfindx = np.where(self.fft_freqs < 0.4)[0]
            cal_all_band = np.interp(self.fft_freqs[np.max(minfindx) + 1:np.shape(self.fft_freqs)[0]],
                                     self.calibration_data[:, 0], calt)
            thmag = np.zeros(np.shape(minfindx), )
            thmag[:, ] = 0.2 * self.fft_freqs[0:np.max(minfindx) + 1]
            thph = np.arctan2(32.0, self.fft_freqs[0:np.max(minfindx) + 1])
            th_band = (thmag * np.cos(thph) + (1j * thmag * np.sin(thph))) * 1000
            cal_all_band = np.concatenate((th_band, cal_all_band))
            cal_all_band[0] = 1 + 1j  # To avoid division of zero
            self.calibrated_data = self.xfft / cal_all_band[:, np.newaxis]



"""
GUI of EDI merger and does merging
"""

import numpy as np
import seaborn as sns
import xarray as xr
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication, QPushButton, QVBoxLayout, QWidget, QFileDialog
from matplotlib import pyplot as plt

from sigmt.utils.edi import edi_ops

if hasattr(Qt, 'AA_EnableHighDpiScaling'):
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)

if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)


class EDIMerger(QWidget):
    """
    GUI for EDI merger
    """

    def __init__(self):
        super().__init__()

        self.fig = None
        self.axs = None
        self.scatter1 = None
        self.scatter2 = None
        self.scatter3 = None
        self.scatter4 = None
        self.color1 = None
        self.initial_color1 = None
        self.initial_data = None
        self.scatter_list = None
        self.data = None
        self.edi_cont = None
        self.hmeas = None
        self.components = None
        self.edi_files = None
        self.setWindowTitle("EDI Merger")
        self.setGeometry(100, 100, 250, 200)
        self.setWindowIcon(QIcon(r'sigmt\images\sigmt.ico'))

        # Layout for buttons
        layout = QVBoxLayout()

        self.select_files_button = QPushButton("Select EDI Files")
        self.select_files_button.clicked.connect(self.select_files)
        layout.addWidget(self.select_files_button)

        self.plot_data_button = QPushButton("Plot Data")
        self.plot_data_button.clicked.connect(self.plot_data)
        layout.addWidget(self.plot_data_button)

        self.save_as_edi_button = QPushButton("Save as EDI")
        self.save_as_edi_button.clicked.connect(self.save_as_edi)
        layout.addWidget(self.save_as_edi_button)

        self.close_button = QPushButton("Close")
        self.close_button.clicked.connect(self.close_window)
        layout.addWidget(self.close_button)

        # Set the layout
        self.setLayout(layout)

    def select_files(self) -> None:
        """
        Open file dialog with .edi filter and allow multiple selections
        """
        self.components = [
            'freqs',
            'zxx_r', 'zxx_i', 'zxx_var',
            'zxy_r', 'zxy_i', 'zxy_var',
            'zyx_r', 'zyx_i', 'zyx_var',
            'zyy_r', 'zyy_i', 'zyy_var',
            'tzx_r', 'tzx_i', 'tzx_var',
            'tzy_r', 'tzy_i', 'tzy_var',
            'coh_ex', 'coh_ey'
        ]
        data = {}
        for component in self.components:
            data[component] = []
        #
        options = QFileDialog.Options()
        self.edi_files, _ = QFileDialog.getOpenFileNames(self, "Select EDI Files", "",
                                                         "EDI Files (*.edi);;"
                                                         "All Files (*)", options=options)
        if self.edi_files:
            colorx = self.generate_colors1()
            self.color1 = []
            self.hmeas = []
            cind = 0
            for edi_file in self.edi_files:
                self.edi_cont, data_temp, hmeas_temp = edi_ops.read_edi(edi_file)
                self.hmeas.append(hmeas_temp)
                for component in self.components:
                    data[component].extend(data_temp[component])
                for _ in range(len(data_temp['freqs'])):
                    self.color1.append(colorx[cind])
                cind = cind + 1
            # Rejects components with no data, but accepts if some are NaNs
            filtered_data = {key: value for key, value in data.items()
                             if key == 'freqs' or not np.all(np.isnan(value))}
            self.data = xr.Dataset(
                {key: (['frequency'], filtered_data[key]) for key in filtered_data},
                coords={'frequency': filtered_data['freqs']}
            )
            self.get_res_phase()
            if 5 in self.hmeas:
                indices_of_5 = [index for index, value in enumerate(self.hmeas) if value == 5]
                self.edi_cont, data_temp, hmeas_temp = edi_ops.read_edi(
                    self.edi_files[indices_of_5[0]])

    def plot_data(self) -> None:
        """
        Plots the data from multiple EDIs
        """
        self.fig, self.axs = plt.subplots(2)
        # Plot the scatter points
        self.scatter1 = self.axs[0].scatter(self.data['freqs'], self.data['rho_xy'],
                                            marker="s", color=self.color1,
                                            s=50,
                                            picker=True,
                                            label='XY')  # rho xy
        self.scatter2 = self.axs[0].scatter(self.data['freqs'], self.data['rho_yx'],
                                            marker="o", color=self.color1,
                                            s=50,
                                            picker=True,
                                            label='YX')  # rho yx
        self.scatter3 = self.axs[1].scatter(self.data['freqs'], self.data['phase_xy'],
                                            marker="s", color=self.color1,
                                            s=50,
                                            picker=True)  # phase xy
        self.scatter4 = self.axs[1].scatter(self.data['freqs'], self.data['phase_yx'],
                                            marker="o", color=self.color1,
                                            s=50,
                                            picker=True)  # phase yx

        # Store the scatter plots in a list for easy access
        self.scatter_list = [self.scatter1, self.scatter2, self.scatter3, self.scatter4]
        # Set log scale for scatter1 and scatter2
        self.axs[0].set_xscale('log')
        self.axs[0].set_yscale('log')

        # Add x-axis and y-axis labels
        self.axs[0].set_xlabel('Frequency (Hz)')
        self.axs[0].set_ylabel('App. Res. (Ohm.m.)')

        self.axs[1].set_xlabel('Frequency (Hz)')
        self.axs[1].set_ylabel('Phase (Deg.)')

        # Set log scale for x-axis of scatter3 and scatter4
        self.axs[1].set_xscale('log')
        self.axs[0].grid(which='both', linestyle='-.', linewidth=0.4)
        self.axs[1].grid(which='both', linestyle='-.', linewidth=0.4)
        self.axs[0].invert_xaxis()
        self.axs[1].invert_xaxis()
        self.axs[0].legend(loc='upper right', bbox_to_anchor=(1, 1))

        self.fig.canvas.mpl_connect('pick_event', self.on_pick)
        self.fig.canvas.mpl_connect('button_press_event', self.on_right_click)

        plt.show(block=False)

    def on_pick(self, event):
        """
        Action for left click
        """
        # Get the index of the clicked point
        index = event.ind[0]

        # Determine which scatter plot was clicked
        for scatter in self.scatter_list:
            if event.artist == scatter:
                # Store initial state for undo
                self.initial_data = self.data.copy()
                self.initial_color1 = self.color1.copy()

                # Deleting...
                self.data = self.data.isel(
                    frequency=[i for i in range(len(self.data['frequency'])) if i != index])
                del self.color1[index]

                # Update the scatter plots with the new data
                self.scatter1.set_offsets(
                    np.column_stack((self.data['freqs'], self.data['rho_xy'])))
                self.scatter2.set_offsets(
                    np.column_stack((self.data['freqs'], self.data['rho_yx'])))
                self.scatter3.set_offsets(
                    np.column_stack((self.data['freqs'], self.data['phase_xy'])))
                self.scatter4.set_offsets(
                    np.column_stack((self.data['freqs'], self.data['phase_yx'])))

                # Update the colors
                self.scatter1.set_color(self.color1)
                self.scatter2.set_color(self.color1)
                self.scatter3.set_color(self.color1)
                self.scatter4.set_color(self.color1)

                # Update the plots
                self.axs[0].figure.canvas.draw()

    def on_right_click(self, event):
        """
        Action for right click
        """
        if event.button == 3:  # Right mouse button click
            self.data = self.initial_data.copy()
            self.color1 = self.initial_color1.copy()
            # Restore original scatter plots
            self.scatter1.set_offsets(np.column_stack((self.data['freqs'], self.data['rho_xy'])))
            self.scatter2.set_offsets(np.column_stack((self.data['freqs'], self.data['rho_yx'])))
            self.scatter3.set_offsets(np.column_stack((self.data['freqs'], self.data['phase_xy'])))
            self.scatter4.set_offsets(np.column_stack((self.data['freqs'], self.data['phase_yx'])))

            self.scatter1.set_color(self.color1)
            self.scatter2.set_color(self.color1)
            self.scatter3.set_color(self.color1)
            self.scatter4.set_color(self.color1)
            # Update the plots
            self.axs[0].figure.canvas.draw()

    def save_as_edi(self) -> None:
        """
        Save merged EDI
        """
        data_frame = self.data.to_dataframe().reset_index()
        data_frame = data_frame.sort_values(by=["frequency"], ascending=[False])

        file_formats = "EDI Files (*.edi);;All Files (*)"
        save_path, _ = QFileDialog.getSaveFileName(None, "Save File", "", file_formats)
        if save_path:  # If user selects a path
            with open(save_path, 'w', encoding='utf-8') as file_handle:
                nfindex = 0
                for j in self.edi_cont:
                    if j[:8] == '  NFREQ=':
                        break
                    nfindex += 1
                for i in range(nfindex):
                    file_handle.write(self.edi_cont[i])
                file_handle.write("  NFREQ=" + str(np.shape(self.data['freqs'])[0]) + "\n")
                findex = 0
                for j in self.edi_cont:
                    if j[:5] == '>FREQ':
                        break
                    findex += 1
                for i in range(nfindex + 1, findex - 1):
                    file_handle.write(self.edi_cont[i])

                # ===== FREQ section =====
                file_handle.write("\n")
                file_handle.write(">FREQ //" + str(len(data_frame['frequency'])))
                file_handle.write("\n")

                for i in range(len(data_frame['frequency'])):
                    k = i + 1
                    file_handle.write(f" {data_frame['frequency'][i]:.9E} ")
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                if any(col in data_frame.columns for col in ['zxy_r', 'zyx_r']):
                    # ===== ZROT section =====
                    file_handle.write("\n \n")
                    file_handle.write(">ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        file_handle.write(f" {0:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZXXR section =====
                if 'zxx_r' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZXXR ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zxx_r'][i] < 0:
                            file_handle.write(f"{data_frame['zxx_r'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zxx_r'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZXXI section =====
                if 'zxx_i' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZXXI ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zxx_i'][i] < 0:
                            file_handle.write(f"{data_frame['zxx_i'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zxx_i'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZXX.VAR section =====
                if 'zxx_var' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZXX.VAR ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zxx_var'][i] < 0:
                            file_handle.write(f"{data_frame['zxx_var'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zxx_var'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZXYR section =====
                if 'zxy_r' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZXYR ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zxy_r'][i] < 0:
                            file_handle.write(f"{data_frame['zxy_r'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zxy_r'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZXYI section =====
                if 'zxy_i' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZXYI ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zxy_i'][i] < 0:
                            file_handle.write(f"{data_frame['zxy_i'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zxy_i'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZXY.VAR section =====
                if 'zxy_var' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZXY.VAR ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zxy_var'][i] < 0:
                            file_handle.write(f"{data_frame['zxy_var'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zxy_var'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZYXR section =====
                if 'zyx_r' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZYXR ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zyx_r'][i] < 0:
                            file_handle.write(f"{data_frame['zyx_r'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zyx_r'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZYXI section =====
                if 'zyx_i' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZYXI ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zyx_i'][i] < 0:
                            file_handle.write(f"{data_frame['zyx_i'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zyx_i'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZYX.VAR section =====
                if 'zyx_var' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZYX.VAR ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zyx_var'][i] < 0:
                            file_handle.write(f"{data_frame['zyx_var'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zyx_var'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZYYR section =====
                if 'zyy_r' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZYYR ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zyy_r'][i] < 0:
                            file_handle.write(f"{data_frame['zyy_r'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zyy_r'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZYYI section =====
                if 'zyy_i' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZYYI ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zyy_i'][i] < 0:
                            file_handle.write(f"{data_frame['zyy_i'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zyy_i'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== ZYY.VAR section =====
                if 'zyy_var' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">ZYY.VAR ROT=ZROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['zyy_var'][i] < 0:
                            file_handle.write(f"{data_frame['zyy_var'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['zyy_var'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== TROT section =====
                if any(col in data_frame.columns for col in ['tzx_r', 'tzy_r']):
                    file_handle.write("\n \n")
                    file_handle.write(">TROT.EXP //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        file_handle.write(f" {0:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== TXR.EXP section =====
                if 'tzx_r' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">TXR.EXP ROT=TROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['tzx_r'][i] < 0:
                            file_handle.write(f"{data_frame['tzx_r'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['tzx_r'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== TXI.EXP section =====
                if 'tzx_i' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">TXI.EXP ROT=TROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['tzx_i'][i] < 0:
                            file_handle.write(f"{data_frame['tzx_i'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['tzx_i'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== TXVAR.EXP section =====
                if 'tzx_var' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">TXVAR.EXP ROT=TROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['tzx_var'][i] < 0:
                            file_handle.write(f"{data_frame['tzx_var'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['tzx_var'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== TYR.EXP section =====
                if 'tzy_r' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">TYR.EXP ROT=TROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['tzy_r'][i] < 0:
                            file_handle.write(f"{data_frame['tzy_r'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['tzy_r'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== TYI.EXP section =====
                if 'tzy_i' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">TYI.EXP ROT=TROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['tzy_i'][i] < 0:
                            file_handle.write(f"{data_frame['tzy_i'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['tzy_i'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== TYVAR.EXP section =====
                if 'tzy_var' in data_frame.columns:
                    file_handle.write("\n \n")
                    file_handle.write(">TYVAR.EXP ROT=TROT //" + str(len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        if data_frame['tzy_var'][i] < 0:
                            file_handle.write(f"{data_frame['tzy_var'][i]:.9E} ")
                        else:
                            file_handle.write(f" {data_frame['tzy_var'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== cohEx section =====
                if 'coh_ex' in data_frame.columns:
                    file_handle.write("\n \n")
                    if 5 in self.hmeas:
                        file_handle.write(
                            ">COH  MEAS1=" + str(101.001) +
                            " MEAS2=" + str(107.001) + "  ROT=NORTH //" + str(
                                len(data_frame['frequency'])))
                    else:
                        file_handle.write(
                            ">COH  MEAS1=" + str(101.001) +
                            " MEAS2=" + str(105.001) + "  ROT=NORTH //" + str(
                                len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        file_handle.write(f" {data_frame['coh_ex'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                # ===== cohEy section =====
                if 'coh_ey' in data_frame.columns:
                    file_handle.write("\n \n")
                    if 5 in self.hmeas:
                        file_handle.write(
                            ">COH  MEAS1=" + str(102.001) +
                            " MEAS2=" + str(106.001) + "  ROT=NORTH //" + str(
                                len(data_frame['frequency'])))
                    else:
                        file_handle.write(
                            ">COH  MEAS1=" + str(102.001) +
                            " MEAS2=" + str(104.001) + "  ROT=NORTH //" + str(
                                len(data_frame['frequency'])))
                    file_handle.write("\n")

                    for i in range(len(data_frame['frequency'])):
                        k = i + 1
                        file_handle.write(f" {data_frame['coh_ey'][i]:.9E} ")
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")
                    file_handle.write("\n \n")
                    file_handle.write(">END")
                    file_handle.close()

    def generate_colors1(self) -> None:
        """
        Generate a list of 'num_colors' distinct colors
        """
        return sns.color_palette("dark", len(self.edi_files))

    def generate_colors2(self) -> None:
        """
        Generate a list of 'num_colors' distinct colors
        """
        return sns.color_palette("deep", len(self.edi_files))

    def get_res_phase(self) -> None:
        """
        Calculates resistivity and phase from impedance values
        """
        ftlist = self.data['freqs']
        zxy = self.data['zxy_r'] + 1j * self.data['zxy_i']
        zyx = self.data['zyx_r'] + 1j * self.data['zyx_i']

        self.data['rho_xy'] = (0.2 / ftlist) * (abs(zxy) ** 2)
        self.data['rho_yx'] = (0.2 / ftlist) * (abs(zyx) ** 2)
        self.data['phase_xy'] = np.degrees(np.arctan(zxy.imag / zxy.real))
        self.data['phase_yx'] = np.degrees(np.arctan(zyx.imag / zyx.real))

    def close_window(self) -> None:
        """
        Closes the window
        """
        self.close()



import numpy as np
import xarray as xr


def cohex(bandavg: xr.Dataset) -> np.ndarray:
    """
    Computes coherency values for ex output channel.

    :param bandavg: Dataset containing the auto- and cross- spectra values and
    impedance values for all time windows at all target frequencies.
    :type bandavg: xr.Dataset
    :return: Returns coherency values for all events.
    :rtype: np.ndarray

    """
    # Ex predicted
    zxx = bandavg['zxx']
    zxy = bandavg['zxy']
    zpz = zxx * np.conj(bandavg['exhx']) + zxy * np.conj(bandavg['exhy'])
    zpx = zxx * bandavg['hxhx'] + zxy * bandavg['hyhx']
    zpy = zxx * bandavg['hxhy'] + zxy * bandavg['hyhy']
    zpzp = zxx * np.conj(zpx) + zxy * np.conj(zpy)
    zpzp = zpzp * bandavg['exex']

    coh = np.where(np.abs(zpzp) > 0, zpz / np.sqrt(zpzp), 1 + 1j)
    coh = np.abs(coh)

    coh = np.where(coh > 1.0, 1 / coh, coh)
    return coh


def cohey(bandavg: xr.Dataset) -> np.ndarray:
    """
    Computes coherency values for ey output channel.

    :param bandavg: Dataset containing the auto- and cross- spectra values and
    impedance values for all time windows at all target frequencies.
    :type bandavg: xr.Dataset
    :return: Returns coherency values for all events.
    :rtype: np.ndarray

    """
    # Ey predicted
    zyx = bandavg['zyx']
    zyy = bandavg['zyy']
    zpz = zyx * np.conj(bandavg['eyhx']) + zyy * np.conj(bandavg['eyhy'])
    zpx = zyx * bandavg['hxhx'] + zyy * bandavg['hyhx']
    zpy = zyx * bandavg['hxhy'] + zyy * bandavg['hyhy']
    zpzp = zyx * np.conj(zpx) + zyy * np.conj(zpy)
    zpzp = zpzp * bandavg['eyey']

    coh = np.where(np.abs(zpzp) > 0, zpz / np.sqrt(zpzp), 1 + 1j)
    coh = np.abs(coh)

    coh = np.where(coh > 1.0, 1 / coh, coh)
    return coh


def cohhz(bandavg: xr.Dataset) -> np.ndarray:
    """
    Computes coherency values for hz output channel.

    :param bandavg: Dataset containing the auto- and cross- spectra values and
    impedance values for all time windows at all target frequencies.
    :type bandavg: xr.Dataset
    :return: Returns coherency values for all events.
    :rtype: np.ndarray

    """
    # Ex predicted
    tzx = bandavg['tzx']
    tzy = bandavg['tzy']
    zpz = tzx * np.conj(bandavg['hzhx']) + tzy * np.conj(bandavg['hzhy'])
    zpx = tzx * bandavg['hxhx'] + tzy * bandavg['hyhx']
    zpy = tzx * bandavg['hxhy'] + tzy * bandavg['hyhy']
    zpzp = tzx * np.conj(zpx) + tzy * np.conj(zpy)
    zpzp = zpzp * bandavg['hzhz']

    coh = np.where(np.abs(zpzp) > 0, zpz / np.sqrt(zpzp), 1 + 1j)
    coh = np.abs(coh)

    coh = np.where(coh > 1.0, 1 / coh, coh)
    return coh


def pdvalues(bandavg: xr.Dataset) -> tuple:
    """
    Computes polarization directions for electric and magnetic field.

    :param bandavg: Dataset containing the auto- and cross- spectra values and
    impedance values for all time windows at all target frequencies.
    :type bandavg: xr.Dataset
    :return: Tuple containing polarization directions for electric field and magnetic field.
    :rtype: tuple

    """
    hxhx = bandavg['hxhx']
    hyhy = bandavg['hyhy']
    hxhy = bandavg['hxhy']
    alpha_h = np.arctan(2 * np.real(hxhy) / (hxhx - hyhy))
    alpha_h_deg = np.degrees(np.real(alpha_h))

    exex = bandavg['exex']
    eyey = bandavg['eyey']
    exey = bandavg['exey']
    alpha_e = np.arctan(2 * np.real(exey) / (exex - eyey))
    alpha_e_deg = np.degrees(np.real(alpha_e))
    return alpha_h_deg, alpha_e_deg
